from typing import Iterable, Optional

import click
from requests import HTTPError

from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core import ui
from ggshield.core.errors import UnexpectedError
from ggshield.core.text_utils import pluralize
from ggshield.verticals.hmsl import Secret, get_client
from ggshield.verticals.hmsl.collection import PreparedSecrets
from ggshield.verticals.hmsl.output import show_results


def check_secrets(
    ctx: click.Context,
    prepared_secrets: PreparedSecrets,
    full_hashes: bool,
):
    """
    Common code to check secrets and display results for check commands.
    """
    # Query the API
    ui.display_info("Querying HasMySecretLeaked...")
    ctx_obj = ContextObj.get(ctx)
    client = get_client(ctx_obj.config, hmsl_command_path=ctx.command_path)
    found: Iterable[Secret] = []
    error: Optional[Exception] = None
    try:
        found = list(client.check(prepared_secrets.payload, full_hashes=full_hashes))
    except (ValueError, HTTPError) as exception:
        error = exception
    ui.display_info(
        f"{client.quota.remaining} {pluralize('credit', client.quota.remaining)} left for today."
    )

    # Display results and error
    show_results(
        found, prepared_secrets.mapping, json_output=ctx_obj.use_json, error=error
    )
    if error:
        raise UnexpectedError(str(error))



import os
import subprocess
from pathlib import Path
from typing import Any, List

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.hooks import check_user_requested_skip
from ggshield.core import ui
from ggshield.core.client import create_client_from_config
from ggshield.core.scan import Commit, ScanContext, ScanMode
from ggshield.core.scanner_ui import create_scanner_ui
from ggshield.utils.git_shell import check_git_dir, git
from ggshield.verticals.secret import SecretScanCollection, SecretScanner
from ggshield.verticals.secret.output import SecretTextOutputHandler


def check_is_merge_without_conflict() -> bool:
    """Check if the reflog action is a merge without conflict"""
    return os.getenv("GIT_REFLOG_ACTION", "").split(" ")[0] == "merge"


def get_merge_branch_from_reflog() -> str:
    """Get the branch that was merged from the reflog"""
    return os.getenv("GIT_REFLOG_ACTION", "").split(" ")[-1]


def check_is_merge_with_conflict(cwd: Path) -> bool:
    """Check if MERGE_HEAD exists  (meaning we are in a merge with conflict)"""
    try:
        git(["rev-parse", "--verify", "-q", "MERGE_HEAD"], cwd=cwd)
        # MERGE_HEAD exists
        return True
    except subprocess.CalledProcessError:
        # MERGE_HEAD does not exist
        return False


@click.command()
@click.option(
    "--scan-all-merge-files",
    is_flag=True,
    help="When scanning a merge commit, scan all files, including those that merged without conflicts.",
)
@click.argument("precommit_args", nargs=-1, type=click.UNPROCESSED)
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def precommit_cmd(
    ctx: click.Context,
    scan_all_merge_files: bool,
    precommit_args: List[str],
    **kwargs: Any,
) -> int:  # pragma: no cover
    """
    Scan as a pre-commit hook all changes that have been staged in a git repository.
    """
    ctx_obj = ContextObj.get(ctx)
    ctx_obj.client = create_client_from_config(ctx_obj.config)
    config = ctx_obj.config

    if check_user_requested_skip():
        return 0

    output_handler = SecretTextOutputHandler(
        verbose=ui.is_verbose(),
        client=ctx_obj.client,
        output=None,
        secret_config=config.user_config.secret,
    )
    check_git_dir()

    scan_context = ScanContext(
        scan_mode=ScanMode.PRE_COMMIT,
        command_path=ctx.command_path,
        target_path=Path.cwd(),
    )

    # Get the commit object
    if not scan_all_merge_files and check_is_merge_with_conflict(Path.cwd()):
        commit = Commit.from_merge(ctx_obj.exclusion_regexes)
    elif not scan_all_merge_files and check_is_merge_without_conflict():
        merge_branch = get_merge_branch_from_reflog()
        commit = Commit.from_merge(ctx_obj.exclusion_regexes, merge_branch)
    else:
        commit = Commit.from_staged(ctx_obj.exclusion_regexes)

    scanner = SecretScanner(
        client=ctx_obj.client,
        cache=ctx_obj.cache,
        scan_context=scan_context,
        secret_config=config.user_config.secret,
    )
    with create_scanner_ui(len(commit.urls)) as scanner_ui:
        results = scanner.scan(commit.get_files(), scanner_ui)

    return_code = output_handler.process_scan(
        SecretScanCollection(id="cached", type="pre-commit", results=results)
    )
    if return_code:
        ui.display_info(ctx_obj.client.remediation_messages.pre_commit)
    return return_code



import json
from unittest import mock

import jsonschema
import pytest
from pygitguardian.models import HealthCheckResponse
from pytest_voluptuous import S
from voluptuous.validators import All, In, Match

from ggshield.__main__ import cli
from ggshield.core.config.config import ConfigSource
from ggshield.utils.os import cd
from tests.unit.conftest import assert_invoke_ok, my_vcr


def test_quota(cli_fs_runner, quota_json_schema):
    with my_vcr.use_cassette("quota"):
        cmd = ["quota", "--json"]
        cli_fs_runner.mix_stderr = False
        result = cli_fs_runner.invoke(cli, cmd, color=False)
        assert_invoke_ok(result)

    dct = json.loads(result.output)
    jsonschema.validate(dct, quota_json_schema)

    assert dct["count"] + dct["remaining"] == dct["limit"]


def test_api_status(cli_fs_runner, api_status_json_schema):
    with my_vcr.use_cassette("test_health_check"):
        cmd = ["api-status", "--json"]
        cli_fs_runner.mix_stderr = False
        result = cli_fs_runner.invoke(cli, cmd, color=False)
        assert_invoke_ok(result)

    dct = json.loads(result.output)
    jsonschema.validate(dct, api_status_json_schema)

    assert (
        S(
            All(
                {
                    "detail": "Valid API key.",
                    "instance": Match(r"https://[^\s]+"),
                    "status_code": 200,
                    "app_version": Match(r"v\d\.\d{1,3}\.\d{1,2}(-rc\.\d)?"),
                    "secrets_engine_version": Match(r"\d\.\d{1,3}\.\d"),
                    "instance_source": In(x.name for x in ConfigSource),
                    "api_key_source": In(x.name for x in ConfigSource),
                }
            )
        )
        == dct
    )


@mock.patch(
    "ggshield.core.config.auth_config.AuthConfig.get_instance_token",
    return_value="token",
)
@mock.patch(
    "pygitguardian.GGClient.health_check",
    return_value=HealthCheckResponse(detail="", status_code=200),
)
def test_api_status_sources(_, hs_mock, cli_fs_runner, tmp_path, monkeypatch):
    """
    GIVEN an api_key and an instance configured anywhere
    WHEN running the api-status command
    THEN the correct api key and instance source are returned
    """
    (tmp_path / ".env").touch()

    monkeypatch.delenv("GITGUARDIAN_INSTANCE", raising=False)
    monkeypatch.delenv("GITGUARDIAN_API_URL", raising=False)

    def get_api_status(env, instance=None):
        with cd(tmp_path):
            cmd = ["api-status", "--json"]
            cli_fs_runner.mix_stderr = False
            if instance:
                cmd.extend(["--instance", instance])
            result = cli_fs_runner.invoke(cli, cmd, color=False, env=env)

        json_res = json.loads(result.output)
        return json_res["instance_source"], json_res["api_key_source"]

    env: dict[str, str | None] = {
        "GITGUARDIAN_INSTANCE": None,
        "GITGUARDIAN_URL": None,
        "GITGUARDIAN_API_KEY": None,
    }
    instance_source, api_key_source = get_api_status(env)
    assert instance_source == ConfigSource.DEFAULT.name
    assert api_key_source == ConfigSource.USER_CONFIG.name

    (tmp_path / ".gitguardian.yaml").write_text(
        "version: 2\ninstance: https://dashboard.gitguardian.com\n"
    )
    instance_source, api_key_source = get_api_status(env)
    assert instance_source == ConfigSource.USER_CONFIG.name
    assert api_key_source == ConfigSource.USER_CONFIG.name

    env["GITGUARDIAN_INSTANCE"] = "https://dashboard.gitguardian.com"
    env["GITGUARDIAN_API_KEY"] = "token"
    instance_source, api_key_source = get_api_status(env)
    assert instance_source == ConfigSource.ENV_VAR.name
    assert api_key_source == ConfigSource.ENV_VAR.name

    (tmp_path / ".env").write_text(
        "GITGUARDIAN_INSTANCE=https://dashboard.gitguardian.com\n"
        "GITGUARDIAN_API_KEY=token"
    )
    instance_source, api_key_source = get_api_status(env)
    assert instance_source == ConfigSource.DOTENV.name
    assert api_key_source == ConfigSource.DOTENV.name

    assert (
        get_api_status(env, instance="https://dashboard.gitguardian.com")[0]
        == ConfigSource.CMD_OPTION.name
    )


@pytest.mark.parametrize("verify", [True, False])
def test_ssl_verify(cli_fs_runner, verify):
    cmd = ["api-status"] if verify else ["--allow-self-signed", "api-status"]

    with mock.patch("ggshield.core.client.GGClient") as client_mock:
        cli_fs_runner.invoke(cli, cmd)
        _, kwargs = client_mock.call_args
        assert kwargs["session"].verify == verify


@pytest.mark.parametrize("command", ["api-status", "quota"])
def test_instance_option(cli_fs_runner, command):
    """
    GIVEN an instance url
    WHEN running a command and passing the instance url as option
    THEN the call resulting from the command is made to the instance url
    """

    uri = "https://dashboard.my-instance.com"

    with mock.patch("ggshield.core.client.GGClient") as client_mock:
        cli_fs_runner.invoke(cli, [command, "--instance", uri])
        _, kwargs = client_mock.call_args
        assert kwargs["base_uri"] == "https://dashboard.my-instance.com/exposed"



from pathlib import Path

import pytest
from click.testing import CliRunner

from ggshield.__main__ import cli
from ggshield.cmd.hmsl.decrypt import load_mapping
from ggshield.verticals.hmsl.client import PREFIX_LENGTH
from ggshield.verticals.hmsl.crypto import hash_string
from tests.unit.conftest import assert_invoke_exited_with, assert_invoke_ok


ENV_FILE = r"""
FOO=bar
PASSWORD=1234
# Don't take "export" into account
export SECRET=foo

this line does not contain a secret

# Commented secrets should be detected
# BAR=prod123
# But comments should be ignored
BAZ=spam # eggs
# All at once
   # FOOBAR=toto42 # a super secret

# Quotes should be ignored
PASSWORD2="123v4@!,asTd"  # ggignore
export PASSWORD3="P@ssw0rd3"  # ggignore
export PASSWORD4="secret with spaces"
export PASSWORD5="escaped \" quote"
PASSWORD6="another escaped \" quote" # and "quotes" in comments afterwards
# Empty secret
PASSWORD7=""

# Ignored common values
KEY1=1
KEY2=0
KEY3=true
KEY4=false
KEY5=on
KEY6=off
KEY7=yes
KEY8=no
KEY9=enabled
KEY10=disabled
KEY11=none
KEY12=null

# Ignored keys
HOST=123.4.56.7
PORT=8080

# Be permissive on case and spaces
 a_secret = its_value

# A last one just in case
LAST=aBcD1234!@#$
"""


@pytest.fixture
def secrets():
    """Prepare a file with some secrets"""
    return ["foo", "bar", "password", "1234"]


@pytest.fixture
def secrets_path(secrets, tmp_path: Path):
    secrets_path = tmp_path / "secrets.txt"
    # add an empty line
    lines = secrets[:2] + [""] + secrets[2:]
    secrets_path.write_text("\n".join(lines))
    return secrets_path


@pytest.fixture
def expected_from_env_file(cli_fs_runner):
    with open(".env", "w") as f:
        f.write(ENV_FILE)
    return {
        "FOO": "bar",
        "PASSWORD": "1234",
        "SECRET": "foo",
        "BAZ": "spam",
        "PASSWORD2": "123v4@!,asTd",  # ggignore
        "PASSWORD3": "P@ssw0rd3",  # ggignore
        "PASSWORD4": "secret with spaces",
        "PASSWORD5": 'escaped " quote',
        "PASSWORD6": 'another escaped " quote',
        "a_secret": "its_value",
        "LAST": "aBcD1234!@#$",
    }


def test_hmsl_fingerprint_no_file(cli_fs_runner: CliRunner) -> None:
    """
    GIVEN some secrets
    WHEN running the fingerprint command with no argument
    THEN the return code is 2
    """
    result = cli_fs_runner.invoke(cli, ["hmsl", "fingerprint"])
    assert_invoke_exited_with(result, 2)


def test_hmsl_fingerprint_no_file_2(cli_fs_runner: CliRunner) -> None:
    """
    GIVEN some secrets
    WHEN running the fingerprint command with a non-existing file
    THEN the return code is 2
    """
    result = cli_fs_runner.invoke(cli, ["hmsl", "fingerprint" "none.txt"])
    assert_invoke_exited_with(result, 2)


def test_hmsl_fingerprint_default_behavior(
    cli_fs_runner: CliRunner, secrets, secrets_path
) -> None:
    """
    GIVEN some secrets
    WHEN running the fingerprint command on a file
    THEN the secrets are correctly prepared
    """
    result = cli_fs_runner.invoke(cli, ["hmsl", "fingerprint", str(secrets_path)])
    assert_invoke_ok(result)
    # Payload is a set of prefixes
    prepared = set(open("payload.txt").read().split("\n")) - {""}
    assert {hash_string(secret)[:PREFIX_LENGTH] for secret in secrets} == prepared
    # Mapping contains the full hashes and hints
    raw_mapping = set(open("mapping.txt").read().split("\n")) - {""}
    mapping = {line.partition(":")[0]: line.partition(":")[2] for line in raw_mapping}
    assert {hash_string(secret) for secret in secrets} == mapping.keys()
    assert all("*" in name for name in mapping.values())


def test_hmsl_fingerprint_full_hashes(
    cli_fs_runner: CliRunner, secrets, secrets_path
) -> None:
    """
    GIVEN some secrets
    WHEN running the fingerprint command on a file
         with the --full-hashes flag
    THEN the payload contains hashes
    """
    result = cli_fs_runner.invoke(cli, ["hmsl", "fingerprint", "-f", str(secrets_path)])
    assert_invoke_ok(result)
    # Payload is a set of prefixes
    prepared = set(open("payload.txt").read().split("\n")) - {""}
    assert {hash_string(secret) for secret in secrets} == prepared


def test_hmsl_fingerprint_cleartext(
    cli_fs_runner: CliRunner, secrets, secrets_path
) -> None:
    """
    GIVEN some secrets
    WHEN running the fingerprint command on a file
         with a cleartext strategy
    THEN the mapping contains the secrets
    """
    result = cli_fs_runner.invoke(
        cli, ["hmsl", "fingerprint", "-n", "cleartext", str(secrets_path)]
    )
    assert_invoke_ok(result)
    # Payload is a set of prefixes
    mapping = set(open("mapping.txt").read().split("\n")) - {""}
    names = {line.partition(":")[2] for line in mapping}
    assert names == set(secrets)


def test_hmsl_fingerprint_none(cli_fs_runner: CliRunner, secrets, secrets_path) -> None:
    """
    GIVEN some secrets
    WHEN running the fingerprint command on a file
         with the none strategy
    THEN the mapping only contains hashes
    """
    result = cli_fs_runner.invoke(
        cli, ["hmsl", "fingerprint", "-n", "none", str(secrets_path)]
    )
    assert_invoke_ok(result)
    # Payload is a set of prefixes
    assert ":" not in open("mapping.txt").read()


@pytest.mark.parametrize(
    "prefix, expected", [("toto", "toto-"), ("toto-", "toto-"), ("", "")]
)
def test_hmsl_fingerprint_prefix(
    cli_fs_runner: CliRunner, secrets_path, prefix, expected
) -> None:
    """
    GIVEN some secrets
    WHEN running the fingerprint command on a file
         with a prefix
    THEN the output files have the proper name
    """
    result = cli_fs_runner.invoke(
        cli, ["hmsl", "fingerprint", "-p", prefix, str(secrets_path)]
    )
    assert_invoke_ok(result)
    assert Path(f"{expected}payload.txt").exists()


def test_hmsl_fingerprint_env_file(
    cli_fs_runner: CliRunner, expected_from_env_file
) -> None:
    """
    GIVEN a .env file
    WHEN running the fingerprint command on it
    THEN the keys and secrets are properly parsed
    """
    result = cli_fs_runner.invoke(
        cli, ["hmsl", "fingerprint", "-f", "-t", "env", ".env"]
    )
    assert_invoke_ok(result)
    expected_mapping = {hash_string(v): k for k, v in expected_from_env_file.items()}
    mapping = load_mapping(open("mapping.txt"))
    assert mapping == expected_mapping


def test_hmsl_fingerprint_env_file_cleartext(
    cli_fs_runner: CliRunner, expected_from_env_file
) -> None:
    """
    GIVEN a .env file
    WHEN using a different naming strategy
    THEN the mapping is written as expected
    """
    result = cli_fs_runner.invoke(
        cli, ["hmsl", "fingerprint", "-f", "-t", "env", "-n", "cleartext", ".env"]
    )
    assert_invoke_ok(result)
    expected_mapping = {
        hash_string(secret): secret for secret in expected_from_env_file.values()
    }
    mapping = load_mapping(open("mapping.txt"))
    assert mapping == expected_mapping



from pathlib import Path
from typing import Dict, Optional, Union

from ggshield.core import ui
from ggshield.utils.git_shell import get_remotes

from .supported_ci import SupportedCI


# Note: this does not exist (yet ?) for CircleCI, see
# https://circleci.canny.io/config/p/provide-env-variable-for-branch-name-targeted-by-pull-request
CI_TARGET_BRANCH_ASSOC: Dict[SupportedCI, str] = {
    SupportedCI.GITHUB: "GITHUB_BASE_REF",
    SupportedCI.GITLAB: "CI_MERGE_REQUEST_TARGET_BRANCH_NAME",
    SupportedCI.JENKINS: "CHANGE_TARGET",
    SupportedCI.AZURE: "SYSTEM_PULLREQUEST_TARGETBRANCHNAME",
    SupportedCI.BITBUCKET: "BITBUCKET_PR_DESTINATION_BRANCH",
    SupportedCI.DRONE: "DRONE_COMMIT_BRANCH",
}


def get_remote_prefix(wd: Optional[Union[str, Path]] = None) -> str:
    remotes = get_remotes(wd=wd)
    if len(remotes) == 0:
        # note: this should not happen in practice, esp. in a CI job
        ui.display_verbose("\tNo remote found.")
        return ""
    else:
        ui.display_verbose(f"\tUsing first remote {remotes[0]}.")
        return f"{remotes[0]}/"



import logging
from typing import Any

import click

from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.quota import format_quota_color
from ggshield.verticals.hmsl import get_client


logger = logging.getLogger(__name__)


@click.command()
@click.pass_context
@add_common_options()
def quota_cmd(
    ctx: click.Context,
    **kwargs: Any,
) -> int:
    """
    Get the number of remaining credits for today.
    """

    # Get our client
    config = ContextObj.get(ctx).config
    client = get_client(config, ctx.command_path)

    click.echo(
        f"Quota limit: {client.quota.limit}\n"
        f"Quota available: {format_quota_color(client.quota.remaining, client.quota.limit)}\n"
        f"Quota reset: {client.quota.reset.isoformat()}\n"
    )

    return 0



import dataclasses
from typing import Any, Dict, List, cast

from pygitguardian.client import VERSIONS
from pygitguardian.models import SecretIncident

from ggshield.verticals.secret.extended_match import ExtendedMatch

from ..secret_scan_collection import (
    Error,
    Result,
    Secret,
    SecretScanCollection,
    group_secrets_by_ignore_sha,
)
from .schemas import JSONScanCollectionSchema
from .secret_output_handler import SecretOutputHandler


class SecretJSONOutputHandler(SecretOutputHandler):
    def create_scan_dict(
        self,
        scan: SecretScanCollection,
        incident_details: Dict[str, SecretIncident],
        top: bool = True,
    ) -> Dict[str, Any]:
        scan_dict: Dict[str, Any] = {
            "id": scan.id,
            "type": scan.type,
            "total_incidents": 0,
            "total_occurrences": 0,
        }
        if scan.extra_info:
            scan_dict["extra_info"] = scan.extra_info

        if top:
            scan_dict["secrets_engine_version"] = VERSIONS.secrets_engine_version

        if scan.results:
            for result in scan.results.results:
                result_dict = self.process_result(result, incident_details)
                scan_dict.setdefault("results", []).append(result_dict)
                scan_dict["total_incidents"] += result_dict["total_incidents"]
                scan_dict["total_occurrences"] += result_dict["total_occurrences"]

            for error in scan.results.errors:
                error_dict = self.process_error(error)
                scan_dict.setdefault("errors", []).append(error_dict)

        if scan.scans:
            for inner_scan in scan.scans_with_results:
                inner_scan_dict = self.create_scan_dict(
                    inner_scan, top=False, incident_details=incident_details
                )
                scan_dict.setdefault("scans", []).append(inner_scan_dict)
                scan_dict["total_incidents"] += inner_scan_dict["total_incidents"]
                scan_dict["total_occurrences"] += inner_scan_dict["total_occurrences"]
        return scan_dict

    def _process_scan_impl(self, scan: SecretScanCollection) -> str:
        if self.with_incident_details:
            assert self.client
            incident_details = scan.get_incident_details(self.client)
        else:
            incident_details = {}
        scan_dict = self.create_scan_dict(
            scan, top=True, incident_details=incident_details
        )
        text = JSONScanCollectionSchema().dumps(scan_dict)
        # dumps() return type is not defined, so cast `text`, otherwise mypy complains
        return cast(str, text)

    def process_result(
        self, result: Result, incident_details: Dict[str, SecretIncident]
    ) -> Dict[str, Any]:
        result_dict: Dict[str, Any] = {
            "filename": result.path,
            "mode": result.filemode.name,
            "incidents": [],
            "total_occurrences": 0,
            "total_incidents": 0,
        }
        sha_dict = group_secrets_by_ignore_sha(result.secrets)
        result_dict["total_incidents"] = len(sha_dict)

        if not self.show_secrets:
            result.censor()

        for ignore_sha, secrets in sha_dict.items():
            flattened_dict = self.serialized_secret(
                ignore_sha, secrets, incident_details
            )
            result_dict["incidents"].append(flattened_dict)
            result_dict["total_occurrences"] += flattened_dict["total_occurrences"]
        return result_dict

    @staticmethod
    def process_error(error: Error) -> Dict[str, Any]:
        error_dict: Dict[str, Any] = {
            "files": [
                {
                    "filename": filename,
                    "mode": filemode.name,
                }
                for filename, filemode in error.files
            ],
            "description": error.description,
        }
        return error_dict

    def serialized_secret(
        self,
        ignore_sha: str,
        secrets: List[Secret],
        incident_details: Dict[str, SecretIncident],
    ) -> Dict[str, Any]:
        flattened_dict: Dict[str, Any] = {
            "occurrences": [],
            "ignore_sha": ignore_sha,
            "policy": secrets[0].policy,
            "detector": secrets[0].detector,
            "total_occurrences": len(secrets),
        }

        if secrets[0].validity:
            flattened_dict["validity"] = secrets[0].validity

        if secrets[0].known_secret:
            flattened_dict["known_secret"] = secrets[0].known_secret
            flattened_dict["incident_url"] = secrets[0].incident_url
            assert secrets[0].incident_url
            details = incident_details.get(secrets[0].incident_url)
            if details is not None:
                flattened_dict["incident_details"] = details

        if secrets[0].ignore_reason is not None:
            flattened_dict["ignore_reason"] = dataclasses.asdict(
                secrets[0].ignore_reason
            )

        for secret in secrets:
            flattened_dict["occurrences"].extend(self.serialize_secret_matches(secret))

        return flattened_dict

    def serialize_secret_matches(
        self,
        secret: Secret,
    ) -> List[Dict[str, Any]]:
        """
        Serialize secret matches. The method uses MatchSpan to get the start and
        end index of the match.
        Returns a list of matches.
        """
        matches_list: List[Dict[str, Any]] = []
        for match in secret.matches:
            assert isinstance(match, ExtendedMatch)

            match_dict: Dict[str, Any] = {
                "match": match.match,
                "match_type": match.match_type,
                "line_start": match.line_start,
                "line_end": match.line_end,
                "index_start": match.span.column_index_start,
                "index_end": match.span.column_index_end,
            }
            if match.pre_line_start is not None and match.pre_line_end is not None:
                match_dict["pre_line_start"] = match.pre_line_start
                match_dict["pre_line_end"] = match.pre_line_end
            if match.post_line_start is not None and match.post_line_end is not None:
                match_dict["post_line_start"] = match.post_line_start
                match_dict["post_line_end"] = match.post_line_end
            matches_list.append(match_dict)
        return matches_list



import logging
import time
from typing import Optional, Tuple

import requests

from ggshield import __version__
from ggshield.core.dirs import get_cache_dir

from .config.utils import load_yaml_dict, save_yaml_dict


logger = logging.getLogger(__name__)
CACHE_FILE = get_cache_dir() / "update_check.yaml"

CHECK_AT_KEY = "check-at"


# Use a short timeout to prevent blocking
CHECK_TIMEOUT = 5


def _split_version(version: str) -> Tuple[int, ...]:
    return tuple([int(x) for x in version.split(".")])


def load_last_check_time() -> Optional[float]:
    """Returns the last time we checked, or None if not available"""
    try:
        cached_data = load_yaml_dict(CACHE_FILE)
        if cached_data is None:
            # File does not exist, do not log any warning
            return None
        return float(cached_data[CHECK_AT_KEY])
    except Exception as e:
        logger.warning("Could not load cached latest version: %s", repr(e))
        return None


def save_last_check_time(check_at: float) -> None:
    """Store the last time we checked"""
    save_yaml_dict({CHECK_AT_KEY: check_at}, CACHE_FILE)


def check_for_updates() -> Optional[str]:
    """
    Check for ggshield updates on GitHub. Return the latest version if available.
    Query GitHub API at most once per day and save locally the latest version in a file.
    """
    check_at = load_last_check_time()

    if check_at is not None and (time.time() - check_at < 24 * 60 * 60):
        # We checked today, no need to check again
        return None

    logger.debug("Checking the latest released version of ggshield...")

    # Save check time now so that it is saved even if the check fails. This ensures we
    # don't try for every command if the user does not have network access.
    try:
        save_last_check_time(time.time())
    except Exception as e:
        logger.warning("Could not save time of version check to cache: %s", repr(e))
        # Do not continue if we can't save check time. If we continue we are going to
        # send requests to api.github.com every time ggshield is called.
        return None

    try:
        resp = requests.get(
            "https://api.github.com/repos/GitGuardian/GGShield/releases/latest",
            headers={
                "Accept": "application/vnd.github+json",
                "User-Agent": f"GGShield {__version__}",
                "X-Github-Api-Version": "2022-11-28",
            },
            timeout=CHECK_TIMEOUT,
        )
    except Exception as e:
        logger.warning("Failed to connect to api.github.com: %s", repr(e))
        return None

    if resp.status_code != 200:
        # Handle GitHub rate limit responses gracefully
        # https://docs.github.com/en/rest/overview/resources-in-the-rest-api?apiVersion=2022-11-28#rate-limiting
        if int(resp.headers.get("X-RateLimit-Remaining", -1)) == 0:
            logger.warning("GitHub rate limit exceeded - rescheduling update check")

            # Reset the next update check based on when the GH API quota resets
            check_at = int(resp.headers.get("X-RateLimit-Reset", -1)) - 24 * 60 * 60
            if check_at < 0:
                # Somehow we've hit the rate limit and the reset header is missing
                # This can only happen if GH changes their responses
                logger.warning("Failed rescheduling update check")
                return None

            try:
                save_last_check_time(check_at)
            except Exception as e:
                logger.warning(
                    "Could not save time of version check to cache: %s", repr(e)
                )

            return None

        logger.warning("Failed to check: %s", resp.text)
        return None

    try:
        data = resp.json()
        latest_version: str = data["tag_name"][1:]

        current_version_split = _split_version(__version__)
        latest_version_split = _split_version(latest_version)
    except Exception as e:
        logger.warning("Failed to parse response: %s", repr(e))
        return None

    if current_version_split < latest_version_split:
        return latest_version
    return None



from typing import Any

import click

from ggshield.cmd.hmsl.check_secret_manager.hashicorp_vault import (
    check_hashicorp_vault_cmd,
)
from ggshield.cmd.utils.common_options import add_common_options


@click.group(
    commands={
        "hashicorp-vault": check_hashicorp_vault_cmd,
    },
)
@add_common_options()
def check_secret_manager_group(**kwargs: Any) -> None:
    """Check if secrets from a secret manager have leaked."""
    pass



import os
import threading
import time
from abc import ABC, abstractmethod
from enum import IntEnum, auto
from logging import LogRecord
from typing import Any, NamedTuple

from typing_extensions import Self


class Level(IntEnum):
    ERROR = auto()
    WARNING = auto()
    INFO = auto()
    VERBOSE = auto()
    DEBUG = auto()


NAME_BY_LEVEL = {
    Level.DEBUG: "Debug",
    Level.VERBOSE: "Verbose",
    Level.INFO: "Info",
    Level.WARNING: "Warning",
    Level.ERROR: "Error",
}


class GGShieldProgress(ABC):
    """
    A generic progress bar. Must be created using `GGShieldUI.create_progress()` and
    used as a context manager.
    """

    @abstractmethod
    def advance(self, amount: int) -> None:
        """
        Move the progress bar by `amount` steps.
        """
        ...

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: Any) -> None:
        pass


class GGShieldUI(ABC):
    """Represents GGShield "UI", all progress-feedback code should ultimately go through
    this class.
    """

    def __init__(self):
        self.level = Level.INFO

    @abstractmethod
    def create_progress(self, total: int) -> GGShieldProgress:
        """
        Creates a generic progress bar, not tied not scannables.
        """
        ...

    def display_debug(self, message: str) -> None:
        if self.level >= Level.DEBUG:
            self._echo(Level.DEBUG, message)

    def display_verbose(self, message: str) -> None:
        if self.level >= Level.VERBOSE:
            self._echo(Level.VERBOSE, message)

    def display_heading(self, message: str) -> None:
        if self.level >= Level.INFO:
            self._echo_heading(message)

    def display_info(self, message: str) -> None:
        if self.level >= Level.INFO:
            self._echo(Level.INFO, message)

    def display_warning(self, message: str) -> None:
        if self.level >= Level.WARNING:
            self._echo(Level.WARNING, message)

    def display_error(self, message: str) -> None:
        if self.level >= Level.ERROR:
            self._echo(Level.ERROR, message)

    def log(self, record: LogRecord) -> None:
        """
        Print a log record produced by the logging package.
        Should not be called directly.
        """
        ...

    @abstractmethod
    def _echo(self, level: Level, message: str) -> None:
        """
        Format `message` according to `level` and print it.
        """
        ...

    @abstractmethod
    def _echo_heading(self, message: str) -> None:
        """
        Format `message` as a heading and print it.
        """
        ...


class DebugInfo(NamedTuple):
    """
    Helper class to gather useful information for debugging
    """

    timestamp: str
    process_id: int
    thread_id: int

    @staticmethod
    def create() -> "DebugInfo":
        return DebugInfo(
            timestamp=time.strftime("%Y-%m-%d %H:%M:%S"),
            process_id=os.getpid(),
            thread_id=threading.get_native_id(),
        )



import pytest

from ggshield.core.config.utils import get_global_path


TEST_AUTH_CONFIG = {
    "default_token_lifetime": 2,
    "instances": [
        {
            "name": "default",
            "url": "https://dashboard.gitguardian.com",
            "default_token_lifetime": 1,
            "accounts": [
                {
                    "workspace_id": 23,
                    "token": "62890f237c703c92fbda8236ec2a055ac21332a46115005c976d68b900535fb5",  # ggignore
                    "type": "pat",
                    "token_name": "my_token",
                    "expire_at": "2022-02-23T12:34:56+00:00",
                }
            ],
        },
        {
            "name": None,
            "url": "https://dashboard.onprem.example.com",
            "default_token_lifetime": 0,  # no expiry
            "accounts": [
                {
                    "workspace_id": 1,
                    "token": "8ecffbaeedcd2f090546efeed3bc48a5f4a04a1196637aef6b3f6bbcfd58a96b",  # ggignore
                    "type": "sat",
                    "token_name": "my_other_token",
                    "expire_at": "2022-02-24T12:34:56+00:00",
                }
            ],
        },
    ],
}


@pytest.fixture
def local_config_path():
    yield ".gitguardian"


@pytest.fixture()
def global_config_path():
    yield get_global_path(".gitguardian")



import subprocess
from pathlib import Path
from typing import Callable, List, Tuple

import pytest

from ggshield.core.filter import init_exclusion_regexes
from ggshield.core.scan import Commit
from ggshield.utils.git_shell import Filemode
from tests.conftest import is_windows
from tests.repository import Repository


PATCH_SEPARATION = (
    """commit 3e0d3805080b044ab221fa8b8998e3039be0a5ca6
Author: Testificate Jose <test@test.test>
Date:   Fri Oct 18 13:20:00 2012 +0100
"""
    + ":100644 000000 1233aef 0000000 D\0ggshield/tests/cassettes/test_files_yes.yaml\0"
    + ":000000 100644 0000000 19465ef A\0tests/test_scannable.py\0"
    + ":100644 100755 b4d3aef b4d3aef M\0bin/shutdown.sh\0"
    + ":000000 100644 0000000 12356ef A\0.env\0"
    + ":100644 100644 ac204ec ac204ec R100\0ggshield/tests/test_config.py\0tests/test_config.py\0"
    + ":100644 100644 6546aef b41653f M\0data/utils/email_sender.py\0"
    + """\0diff --git a/ggshield/tests/cassettes/test_files_yes.yaml b/ggshield/tests/cassettes/test_files_yes.yaml
deleted file mode 100644
index 1233aef..0000000
--- a/ggshield/tests/cassettes/test_files_yes.yaml
+++ /dev/null
@@ -1,45 +0,0 @@
-interactions:

diff --git a/tests/test_scannable.py b/tests/test_scannable.py
new file mode 100644
index 0000000..19465ef
--- /dev/null
+++ b/tests/test_scannable.py
@@ -0,0 +1,112 @@
+from collections import namedtuple

diff --git a/bin/shutdown.sh b/bin/shutdown.sh
old mode 100644
new mode 100755

diff --git a/.env b/.env
new file mode 100644
index 0000000..12356ef
--- /dev/null
+++ b/.env
@@ -0,0 +1,112 @@
CHECK_ENVIRONMENT=true

diff --git a/ggshield/tests/test_config.py b/tests/test_config.py
similarity index 100%
rename from ggshield/tests/test_config.py
rename to tests/test_config.py

diff --git a/data/utils/email_sender.py b/data/utils/email_sender.py
index 6546aef..b41653f 100644
--- a/data/utils/email_sender.py
+++ b/data/utils/email_sender.py
@@ -49,6 +49,7 @@ def send_email(config, subject, content, tos, seperate):
    def send_email(subject, content, to, seperate=True):
+   logger.bind(operation_name="send_email")
@@ -73,22 +74,11 @@ def send_email(subject, content, to, seperate=True):
-   removed
+   added
"""
)  # noqa

EXPECTED_PATCH_CONTENT = (
    (
        "ggshield/tests/cassettes/test_files_yes.yaml",
        """@@ -1,45 +0,0 @@
-interactions:

""",
    ),
    (
        "tests/test_scannable.py",
        """@@ -0,0 +1,112 @@
+from collections import namedtuple

""",
    ),
    (
        ".env",
        """@@ -0,0 +1,112 @@
CHECK_ENVIRONMENT=true

""",
    ),
    (
        "data/utils/email_sender.py",
        """@@ -49,6 +49,7 @@ def send_email(config, subject, content, tos, seperate):
    def send_email(subject, content, to, seperate=True):
+   logger.bind(operation_name="send_email")
@@ -73,22 +74,11 @@ def send_email(subject, content, to, seperate=True):
-   removed
+   added
""",
    ),
)


def test_patch_separation():
    commit = Commit.from_patch(PATCH_SEPARATION)
    files = list(commit.get_files())

    assert commit.info.author == "Testificate Jose"
    assert commit.info.email == "test@test.test"
    assert commit.info.date == "Fri Oct 18 13:20:00 2012 +0100"

    assert len(files) == len(EXPECTED_PATCH_CONTENT)
    for file_, (path, content) in zip(files, EXPECTED_PATCH_CONTENT):
        assert file_.path == Path(path)
        assert file_.content == content


def test_patch_separation_ignore():
    file_to_ignore = ".env"
    commit = Commit.from_patch(
        PATCH_SEPARATION, init_exclusion_regexes([file_to_ignore])
    )
    files = list(commit.get_files())

    assert len(files) == 3
    assert not (any(entry.filename == file_to_ignore for entry in files))


def create_file(repo: Repository, name: str = "f", content: str = "Hello\n") -> Path:
    path = repo.path / name
    path.write_text(content)
    return path


def file_append(path: Path, content: str) -> None:
    with path.open("a") as f:
        f.write(content)


def file_search_and_replace(path: Path, src: str, dst: str) -> None:
    content = path.read_text().replace(src, dst)
    path.write_text(content)


def scenario_add(repo: Repository) -> None:
    repo.add(create_file(repo))
    repo.create_commit()


def scenario_add_unusual_chars(repo: Repository) -> None:
    repo.add(create_file(repo, "I'm unusual!"))
    repo.create_commit()


def scenario_add_two(repo: Repository) -> None:
    repo.add(create_file(repo, "one"))
    repo.add(create_file(repo, "two"))
    repo.create_commit()


def scenario_modify(repo: Repository) -> None:
    path = create_file(repo, "f", "Old content")
    repo.add(path)
    repo.create_commit()
    path.write_text("New content")
    repo.add(path)
    repo.create_commit()


def scenario_remove(repo: Repository) -> None:
    path = create_file(repo)
    repo.add(path)
    repo.create_commit()
    repo.git("rm", path)
    repo.create_commit()


def scenario_rename(repo: Repository) -> None:
    path = create_file(repo, "old file.txt")
    repo.add(path)
    repo.create_commit()
    repo.git("mv", path.name, "new file.txt")
    repo.create_commit()


def scenario_chmod(repo: Repository) -> None:
    path = create_file(repo)
    repo.add(path)
    repo.create_commit()
    path.chmod(0o755)
    repo.add(path)
    repo.create_commit()


def scenario_rename_chmod_modify(repo: Repository) -> None:
    path = create_file(repo, "oldscript", "another script\n")
    repo.add(path)
    repo.create_commit()

    # rename
    new_path = repo.path / "newscript"
    repo.git("mv", path.name, new_path.name)

    # chmod
    new_path.chmod(0o755)

    # modify, but not too much so that it's still considered as a rename
    file_append(new_path, "more content\n")

    repo.add(new_path)
    repo.create_commit()


def scenario_merge(repo: Repository) -> None:
    long_file = create_file(
        repo,
        "longfile",
        """Some content

long enough


for changes
to merge

without conflict...

hopefully
""",
    )
    repo.add(long_file)
    repo.create_commit()

    # Append to long_file in branch b1
    repo.create_branch("b1")
    file_append(long_file, "Hello from b1\n")
    repo.add(long_file)
    repo.create_commit()

    # Replace text in long_file in branch b2
    repo.checkout("main")
    repo.create_branch("b2")
    file_search_and_replace(long_file, "for changes", "for changes from b2")
    repo.add(long_file)
    repo.create_commit()

    # Replace text in branch main
    repo.checkout("main")
    file_search_and_replace(long_file, "Some content", "## Some content")
    repo.add(long_file)
    repo.create_commit()

    # Merge all branches
    repo.git("merge", "--no-ff", "b1", "b2")


def scenario_merge_with_changes(repo: Repository) -> None:
    path = create_file(repo, "conflicted")
    repo.add(path)
    repo.create_commit()

    # Append from branch b1
    repo.create_branch("b1")
    file_append(path, "Hello from b1\n")
    repo.add(path)
    repo.create_commit()

    # Append from main
    repo.checkout("main")
    file_append(path, "Hello from main\n")
    repo.add(path)
    repo.create_commit()

    # Try to merge
    try:
        repo.git("merge", "--no-ff", "b1", "--no-commit")
    except subprocess.CalledProcessError:
        pass

    # Solve conflict
    path.write_text("Solve conflict\n")
    repo.add(path)
    repo.create_commit()


def scenario_type_change(repo: Repository) -> None:
    path = create_file(repo)

    # Create `f2` as a symlink to path
    f2_path = repo.path / "f2"
    f2_path.symlink_to(path)
    repo.add(path, f2_path)
    repo.create_commit()

    # Turn `f2` into a file
    f2_path.unlink()
    f2_path.write_text("file")
    repo.add(f2_path)
    repo.create_commit()


@pytest.mark.parametrize(
    ("scenario", "expected_paths_and_modes"),
    [
        pytest.param(*x, id=x[0].__name__)
        for x in [
            (scenario_add, [("f", Filemode.NEW)]),
            (scenario_add_unusual_chars, [("I'm unusual!", Filemode.NEW)]),
            (
                scenario_add_two,
                [
                    ("one", Filemode.NEW),
                    ("two", Filemode.NEW),
                ],
            ),
            (
                scenario_modify,
                [
                    ("f", Filemode.MODIFY),
                ],
            ),
            (
                scenario_remove,
                [
                    ("f", Filemode.DELETE),
                ],
            ),
            (
                scenario_rename,
                [],  # a rename with no content change yields no content
            ),
            (
                scenario_chmod,
                [],  # a permission change with no content change yields no content
            ),
            (
                scenario_rename_chmod_modify,
                [
                    ("newscript", Filemode.RENAME),
                ],
            ),
            (
                scenario_merge,
                [],  # no conflict -> nothing to scan
            ),
            (
                scenario_merge_with_changes,
                [
                    ("conflicted", Filemode.MODIFY),
                ],
            ),
            (
                scenario_type_change,
                [
                    ("f2", Filemode.NEW),
                    ("f2", Filemode.NEW),
                ],
            ),
        ]
    ],
)
def test_from_sha(
    tmp_path,
    scenario: Callable[[Repository], None],
    expected_paths_and_modes: List[Tuple[str, Filemode]],
):
    """
    GIVEN a Commit created from `scenario`
    WHEN Commit.get_files() is called
    THEN it returns files with correct names and modes
    """
    if is_windows() and scenario == scenario_type_change:
        # Path.symlink_to() does not produce the expected diff changes on Windows, so
        # skip this test for now
        pytest.skip()
    repo = Repository.create(tmp_path)
    scenario(repo)

    sha = repo.get_top_sha()
    commit = Commit.from_sha(sha, cwd=tmp_path)
    files = list(commit.get_files())

    paths_and_modes = [(x.path, x.filemode) for x in files]
    assert paths_and_modes == [
        (Path(name), mode) for name, mode in expected_paths_and_modes
    ]


def test_from_sha_gets_right_content_for_conflicts(tmp_path):
    """
    GIVEN a merge commit with a conflict, loaded with Commit.from_sha()
    WHEN Commit.get_files() is called
    THEN it returns the right content
    """
    repo = Repository.create(tmp_path)
    scenario_merge_with_changes(repo)

    sha = repo.get_top_sha()
    commit = Commit.from_sha(sha, cwd=tmp_path)

    files = list(commit.get_files())
    assert len(files) == 1
    content = files[0].content

    # Content has been turned into a single-parent diff
    assert (
        content
        == """
@@ -1,2 +1,1 @@
-Hello
-Hello from main
+Solve conflict
""".strip()
    )


def test_from_sha_files_matches_content(tmp_path):
    """
    GIVEN a commit with many files
    WHEN Commit.get_files() is called
    THEN the reported file names match their expected content
    """
    repo = Repository.create(tmp_path)

    for idx in range(50):
        path = tmp_path / str(idx)
        path.parent.mkdir(exist_ok=True)
        path.write_text(f"{idx}\n")
        repo.add(path)
    repo.create_commit()

    sha = repo.get_top_sha()
    commit = Commit.from_sha(sha, cwd=tmp_path)
    files = list(commit.get_files())

    for file in files:
        last_line = file.content.splitlines()[-1]
        assert last_line == f"+{file.path.name}"


def test_from_staged(tmp_path):
    """
    GIVEN a new file added with `git add`
    AND a Commit instance created from git staged files
    WHEN Commit.get_files() is called
    THEN it returns files with correct names and modes
    """
    repository = Repository.create(tmp_path)
    new_file = tmp_path / "NEW.md"
    new_file.write_text("Hello")
    repository.add(new_file)

    commit = Commit.from_staged(cwd=tmp_path)
    files = list(commit.get_files())

    paths_and_modes = [(x.path, x.filemode) for x in files]
    assert paths_and_modes == [(Path("NEW.md"), Filemode.NEW)]


def test_from_merge(tmp_path):
    """
    GIVEN a Commit instance created from a git merge (after conflict)
    WHEN Commit.get_files() is called
    THEN it returns files with correct names and modes
    """
    repo = Repository.create(tmp_path, initial_branch="master")

    Path(tmp_path / "inital.md").write_text("Initial")
    repo.add(".")
    repo.create_commit("Initial commit on master")

    repo.create_branch("feature_branch")
    repo.checkout("master")
    conflict_file = tmp_path / "conflict.md"
    conflict_file.write_text("Hello")
    Path(tmp_path / "Other.md").write_text("Other")
    repo.add(".")
    repo.create_commit("Commit on master")

    repo.checkout("feature_branch")
    conflict_file.write_text("World")
    Path(tmp_path / "Another.md").write_text("Another")
    repo.add(".")
    repo.create_commit("Commit on feature_branch")

    # Create merge commit with conflict
    with pytest.raises(subprocess.CalledProcessError) as exc:
        repo.git("merge", "master")

    # check stdout for conflict message
    stdout = exc.value.stdout.decode()
    assert "CONFLICT" in stdout

    conflict_file.write_text("Hello World !")
    Path(tmp_path / "new.md").write_text("Something added at conflict")
    repo.add(".")
    commit = Commit.from_merge(cwd=tmp_path)
    files = list(commit.get_files())

    paths_and_modes = [(x.path, x.filemode) for x in files]
    assert paths_and_modes == [
        (Path("conflict.md"), Filemode.MODIFY),
        (Path("new.md"), Filemode.NEW),
    ]


def test_from_merge_filename_with_spaces(tmp_path):
    """
    GIVEN two commits on different branches with a conflict
    involving a filename with spaces
    WHEN Commit.from_merge() is called
    THEN it returns successfully
    AND get_files returns the correct filename
    """
    repo = Repository.create(tmp_path, initial_branch="master")

    Path(tmp_path / "inital.md").write_text("Initial")
    repo.add(".")
    repo.create_commit("Initial commit on master")

    repo.create_branch("feature_branch")
    repo.checkout("master")
    conflict_file = tmp_path / "file with spaces.md"
    conflict_file.write_text("Hello")
    repo.add(".")
    repo.create_commit("Commit on master")

    repo.checkout("feature_branch")
    conflict_file.write_text("World")
    repo.add(".")
    repo.create_commit("Commit on feature_branch")

    # Create merge commit with conflict
    with pytest.raises(subprocess.CalledProcessError):
        repo.git("merge", "master")

    conflict_file.write_text("Hello World !")
    repo.add(".")
    commit = Commit.from_merge(cwd=tmp_path)
    files = list(commit.get_files())
    assert len(files) == 1
    assert files[0].path == Path("file with spaces.md")



from urllib.parse import ParseResult, urlparse

from click import UsageError

from . import ui
from .constants import ON_PREMISE_API_URL_PATH_PREFIX


GITGUARDIAN_DOMAINS = ["gitguardian.com", "gitguardian.tech"]


def clean_url(url: str, warn: bool = False) -> ParseResult:
    """
    Take a dashboard or API URL and removes trailing slashes and useless /v1
    (optionally with a warning).
    """
    parsed_url = urlparse(url)
    if parsed_url.path.endswith("/"):
        parsed_url = parsed_url._replace(path=parsed_url.path[:-1])
    if parsed_url.path.endswith("/v1"):
        parsed_url = parsed_url._replace(path=parsed_url.path[:-3])
        if warn:
            ui.display_warning("Unexpected /v1 path in your URL configuration")
    return parsed_url


def validate_instance_url(url: str, warn: bool = False) -> ParseResult:
    """
    Validate a dashboard URL
    """
    parsed_url = clean_url(url, warn=warn)
    if parsed_url.scheme != "https" and not (
        parsed_url.netloc.startswith("localhost")
        or parsed_url.netloc.startswith("127.0.0.1")
    ):
        raise UsageError(f"Invalid scheme for dashboard URL '{url}', expected HTTPS")
    if any(parsed_url.netloc.endswith("." + domain) for domain in GITGUARDIAN_DOMAINS):
        if parsed_url.path:
            raise UsageError(
                f"Invalid dashboard URL '{url}', got an unexpected path '{parsed_url.path}'"
            )

    return parsed_url


def dashboard_to_api_url(dashboard_url: str, warn: bool = False) -> str:
    """
    Convert a dashboard URL to an API URL.
    handles the SaaS edge case where the host changes instead of the path
    """
    parsed_url = validate_instance_url(dashboard_url, warn=warn)

    if any(parsed_url.netloc.endswith("." + domain) for domain in GITGUARDIAN_DOMAINS):
        parsed_url = parsed_url._replace(
            netloc=parsed_url.netloc.replace("dashboard", "api")
        )
    else:
        parsed_url = parsed_url._replace(
            path=f"{parsed_url.path}{ON_PREMISE_API_URL_PATH_PREFIX}"
        )
    return parsed_url.geturl()


def api_to_dashboard_url(api_url: str, warn: bool = False) -> str:
    """
    Convert an API URL to a dashboard URL.
    handles the SaaS edge case where the host changes instead of the path
    """
    parsed_url = clean_url(api_url, warn=warn)
    if parsed_url.scheme != "https" and not parsed_url.netloc.startswith("localhost"):
        raise UsageError(f"Invalid scheme for API URL '{api_url}', expected HTTPS")
    if any(
        parsed_url.netloc.endswith("." + domain) for domain in GITGUARDIAN_DOMAINS
    ):  # SaaS
        if parsed_url.path:
            raise UsageError(
                f"Invalid API URL '{api_url}', got an unexpected path '{parsed_url.path}'"
            )
        parsed_url = parsed_url._replace(
            netloc=parsed_url.netloc.replace("api", "dashboard")
        )
    elif parsed_url.path.endswith(ON_PREMISE_API_URL_PATH_PREFIX):
        parsed_url = parsed_url._replace(
            path=parsed_url.path[: -len(ON_PREMISE_API_URL_PATH_PREFIX)]
        )
    return parsed_url.geturl()


def urljoin(url: str, *args: str) -> str:
    """
    concatenate each argument with a slash if not already existing.
    unlike urllib.parse.urljoin, this will make sure each element
    is separated by a slash e.g.
    ('http://somesite.com/path1', 'path2') -> http://somesite.com/path1/path2
    ('http://somesite.com/path1/', 'path2') -> http://somesite.com/path1/path2
    ('http://somesite.com/path1', '/path2') -> http://somesite.com/path1/path2
    """
    if url[-1] == "/":
        url = url[:-1]

    for url_part in args:
        if url_part[0] != "/":
            url_part = "/" + url_part
        url += url_part

    return url



import io
import tarfile
import tempfile
from pathlib import Path
from typing import Dict, List, Optional
from unittest.mock import Mock, patch

import click
import pytest

from ggshield.__main__ import cli
from ggshield.core import ui
from ggshield.core.errors import ExitCode
from ggshield.core.git_hooks.ci.commit_range import collect_commit_range_from_ci_env
from ggshield.utils.git_shell import EMPTY_SHA
from ggshield.utils.os import cd
from tests.repository import Repository
from tests.unit.conftest import assert_invoke_exited_with, assert_invoke_ok


@pytest.fixture(autouse=True)
def clear_current_ci_envs(monkeypatch):
    # Make sure the tests are not affected by the fact they themselves are running inside
    # the CI
    monkeypatch.delenv("CI", raising=False)
    monkeypatch.delenv("GITHUB_ACTIONS", raising=False)


@patch("ggshield.core.git_hooks.ci.commit_range.get_list_commit_SHA")
@patch(
    "ggshield.core.git_hooks.ci.commit_range.get_remote_prefix", return_value="origin/"
)
@patch("ggshield.cmd.secret.scan.ci.check_git_dir")
@pytest.mark.parametrize(
    "env,expected_parameter",
    [
        ({"CI_COMMIT_BEFORE_SHA": "before_sha"}, "before_sha~1..HEAD"),
        (
            {
                "CI_COMMIT_BEFORE_SHA": EMPTY_SHA,
                "CI_COMMIT_SHA": "commit_sha",
            },
            "commit_sha~1...",
        ),
        (
            {
                "CI_COMMIT_BEFORE_SHA": EMPTY_SHA,
                "HEAD": "HEAD",
            },
            "HEAD~1...",
        ),
        (
            {
                "CI_COMMIT_BEFORE_SHA": EMPTY_SHA,
                "CI_MERGE_REQUEST_TARGET_BRANCH_NAME": "mr_target_branch_name",
            },
            "origin/mr_target_branch_name..HEAD",
        ),
        (
            {
                "CI_COMMIT_BEFORE_SHA": EMPTY_SHA,
                "CI_MERGE_REQUEST_TARGET_BRANCH_NAME": EMPTY_SHA,
                "CI_COMMIT_SHA": "commit_sha",
            },
            "commit_sha~1...",
        ),
        ({"CI_COMMIT_SHA": "commit_sha"}, "commit_sha~1..."),
        ({"HEAD": "head_sha"}, "HEAD~1..."),
    ],
)
def test_gitlab_ci_range(
    _: Mock,
    get_remote_prefix_mock: Mock,
    get_list_mock: Mock,
    cli_fs_runner: click.testing.CliRunner,
    monkeypatch,
    env: Dict[str, str],
    expected_parameter: str,
    capsys,
):
    """
    GIVEN a GitLab CI environment
    AND verbose mode has been activated
    WHEN gitlab_ci_range() is called
    THEN the correct commit range is requested
    AND stdout is empty (to avoid polluting redirections)
    AND stderr is not empty
    """
    ui.set_level(ui.Level.VERBOSE)

    monkeypatch.setenv("CI", "1")
    monkeypatch.setenv("GITLAB_CI", "1")
    for k, v in env.items():
        monkeypatch.setenv(k, v)

    get_list_mock.return_value = ["a"] * 51

    collect_commit_range_from_ci_env()
    get_list_mock.assert_called_once_with(expected_parameter)

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err != ""


@patch("ggshield.cmd.secret.scan.ci.scan_commit_range")
@patch("ggshield.core.git_hooks.ci.commit_range.get_list_commit_SHA")
@patch("ggshield.cmd.secret.scan.ci.check_git_dir")
@pytest.mark.parametrize(
    ("env", "expected_mode"),
    [
        ({"CI": "true", "GITLAB_CI": "true"}, "ci/GITLAB"),
        ({"CI": "true", "GITHUB_ACTIONS": "true"}, "ci/GITHUB ACTIONS"),
        ({"CI": "true", "TRAVIS": "true"}, "ci/TRAVIS"),
        ({"JENKINS_HOME": "/var/jenkins"}, "ci/JENKINS HOME"),
        (
            {"CI": "true", "JENKINS_URL": "https://ci.example.com"},
            "ci/JENKINS HOME",
        ),  # Can we really have JENKINS_URL but not JENKINS_HOME?
        ({"CI": "true", "CIRCLECI": "true"}, "ci/CIRCLECI"),
        ({"CI": "true", "BITBUCKET_COMMIT": "12345abcd"}, "ci/BITBUCKET PIPELINES"),
        ({"CI": "true", "DRONE": "1", "DRONE_COMMIT_BEFORE": "c0ff331a"}, "ci/DRONE"),
        (
            {"BUILD_BUILDID": "1234", "BUILD_SOURCEVERSION": "b4dd3caf"},
            "ci/AZURE PIPELINES",
        ),
    ],
)
def test_ci_cmd_uses_right_mode_header(
    _: Mock,
    get_list_mock: Mock,
    scan_commit_range_mock: Mock,
    cli_fs_runner: click.testing.CliRunner,
    env: Dict[str, str],
    expected_mode: str,
):
    get_list_mock.return_value = ["a"] * 51
    scan_commit_range_mock.return_value = 0

    # WHEN `secret scan ci` is called
    with patch("os.getenv", env.get):
        result = cli_fs_runner.invoke(cli, ["secret", "scan", "ci"])

    # THEN scan succeeds
    assert_invoke_ok(result)

    # AND scan_commit_range() is called with the right mode
    scan_commit_range_mock.assert_called_once()
    args = scan_commit_range_mock.call_args

    assert args.kwargs["scan_context"].scan_mode == expected_mode


@patch("ggshield.cmd.secret.scan.ci.check_git_dir")
def test_ci_cmd_does_not_work_outside_ci(_, cli_fs_runner: click.testing.CliRunner):
    # GIVEN no CI env
    # WHEN `secret scan ci` is called
    result = cli_fs_runner.invoke(cli, ["secret", "scan", "ci"])

    # THEN it fails
    assert_invoke_exited_with(result, ExitCode.USAGE_ERROR)

    # And the error message explains why
    assert "only be used in a CI environment" in result.stdout


@patch("ggshield.cmd.secret.scan.ci.check_git_dir")
def test_ci_cmd_does_not_work_if_ci_env_is_odd(
    _, monkeypatch, cli_fs_runner: click.testing.CliRunner
):
    # GIVEN an incomplete CI env
    monkeypatch.setenv("CI", "true")

    # WHEN `secret scan ci` is called
    result = cli_fs_runner.invoke(cli, ["secret", "scan", "ci"])

    # THEN it fails
    assert_invoke_exited_with(result, ExitCode.UNEXPECTED_ERROR)

    # And the error message explains why
    assert "Current CI is not detected" in result.stdout


DUMMY_REPO: Optional[tarfile.TarFile] = None


@pytest.fixture
def dummy_repo(tmp_path):
    global DUMMY_REPO
    if not DUMMY_REPO:
        DUMMY_REPO = make_dummy_repo()
    """Return a fresh copy of a dummy repo"""
    DUMMY_REPO.extractall(path=tmp_path)
    return Repository(tmp_path)


def make_dummy_repo():
    """Function to create a dummy repo as a tarfile."""
    result_buffer = io.BytesIO()
    with tempfile.TemporaryDirectory() as tmp_path_str:
        tmp_path = Path(tmp_path_str)
        repo = Repository.create(tmp_path)

        repo.create_commit("initial_commit")

        repo.create_branch("branch")
        repo.create_commit("branch_commit_1")
        repo.create_commit("branch_commit_2")

        repo.checkout("main")
        repo.create_commit("commit_main_1")
        repo.create_commit("commit_main_2")

        repo.create_branch("unrelated_branch")
        repo.create_commit("unrelated_branch_commit")

        result_tar = tarfile.TarFile(fileobj=result_buffer, mode="w")
        result_tar.add(tmp_path_str, arcname="./")
    result_buffer.seek(0)
    return tarfile.TarFile(fileobj=result_buffer, mode="r")


def subjets_for_commits(commit_shas: List[str], dummy_repo: Repository) -> List[str]:
    return [
        dummy_repo.git("show", sha, "--pretty=format:%s", "-s") for sha in commit_shas
    ]


@pytest.mark.parametrize(
    ("env_vars", "expected_subjects"),
    [
        ({"GITHUB_ACTIONS": 1}, ["branch_commit_2"]),
        ({"GITHUB_ACTIONS": 1, "GITHUB_SHA": "branch"}, ["branch_commit_2"]),
        (
            {"GITHUB_ACTIONS": 1, "GITHUB_SHA": "branch~1"},
            ["branch_commit_1", "branch_commit_2"],
        ),
        (
            {
                "GITHUB_ACTIONS": 1,
                "GITHUB_SHA": "branch",
                "GITHUB_DEFAULT_BRANCH": "main",
            },
            ["branch_commit_1", "branch_commit_2"],
        ),
        (
            {
                "GITHUB_ACTIONS": 1,
                "GITHUB_SHA": "branch",
                "GITHUB_DEFAULT_BRANCH": "main",
                "GITHUB_PUSH_BASE_SHA": "branch~1",
            },
            ["branch_commit_2"],
        ),
        (
            {
                "GITHUB_ACTIONS": 1,
                "GITHUB_SHA": "branch",
                "GITHUB_DEFAULT_BRANCH": "main",
                "GITHUB_PUSH_BASE_SHA": "branch~1",
                "GITHUB_BASE_REF": "main",
            },
            ["branch_commit_1", "branch_commit_2"],
        ),
        (
            {
                "GITHUB_ACTIONS": 1,
                "GITHUB_SHA": "branch",
                "GITHUB_DEFAULT_BRANCH": "main",
                "GITHUB_PUSH_BASE_SHA": "branch~1",
                "GITHUB_BASE_REF": EMPTY_SHA,
            },
            ["branch_commit_2"],
        ),
        (
            {
                "GITHUB_ACTIONS": 1,
                "GITHUB_SHA": "branch",
                "GITHUB_DEFAULT_BRANCH": "main",
                "GITHUB_PUSH_BASE_SHA": EMPTY_SHA,
                "GITHUB_BASE_REF": EMPTY_SHA,
            },
            ["branch_commit_1", "branch_commit_2"],
        ),
        (
            {"TRAVIS": 1, "TRAVIS_COMMIT_RANGE": "branch~2..branch"},
            ["branch_commit_1", "branch_commit_2"],
        ),
        (
            {"TRAVIS": 1, "TRAVIS_COMMIT_RANGE": "...branch"},
            ["branch_commit_2"],
        ),
        (
            {
                "GITLAB_CI": 1,
                "CI_COMMIT_SHA": "branch",
                "CI_COMMIT_BEFORE_SHA": "branch~1",
            },
            ["branch_commit_1", "branch_commit_2"],
        ),
    ],
)
def test_collect_commit_range_from_ci_env(
    env_vars,
    expected_subjects,
    monkeypatch,
    cli_fs_runner: click.testing.CliRunner,
    dummy_repo: Repository,
):
    env_vars["CI"] = "1"
    dummy_repo.checkout("branch")

    with patch("os.getenv", env_vars.get), cd(str(dummy_repo.path)):
        commits, _ = collect_commit_range_from_ci_env()

    assert subjets_for_commits(commits, dummy_repo) == expected_subjects



#!/usr/bin/python3
import logging
import multiprocessing
import os
import sys
from io import TextIOWrapper
from pathlib import Path
from typing import Any, List, Optional

import click

from ggshield import __version__
from ggshield.cmd.auth import auth_group
from ggshield.cmd.config import config_group
from ggshield.cmd.hmsl import hmsl_group
from ggshield.cmd.honeytoken import honeytoken_group
from ggshield.cmd.install import install_cmd
from ggshield.cmd.quota import quota_cmd
from ggshield.cmd.secret import secret_group
from ggshield.cmd.secret.scan import scan_group
from ggshield.cmd.status import status_cmd
from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.debug import setup_debug_mode
from ggshield.core import check_updates, ui
from ggshield.core.cache import Cache
from ggshield.core.config import Config
from ggshield.core.env_utils import load_dot_env
from ggshield.core.errors import ExitCode
from ggshield.core.ui import ensure_level, log_utils
from ggshield.core.ui.rich import RichGGShieldUI
from ggshield.utils.click import RealPath
from ggshield.utils.os import getenv_bool


logger = logging.getLogger(__name__)


@scan_group.result_callback()
@click.pass_context
def exit_code(ctx: click.Context, exit_code: int, **kwargs: Any) -> int:
    """
    exit_code guarantees that the return value of a scan is 0
    when exit_zero is enabled
    """
    ctx_obj = ContextObj.get(ctx)
    if (
        exit_code == ExitCode.SCAN_FOUND_PROBLEMS
        and ctx_obj.config.user_config.exit_zero
    ):
        logger.debug("scan exit_code forced to 0")
        sys.exit(ExitCode.SUCCESS)

    logger.debug("scan exit_code=%d", exit_code)
    return exit_code


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    commands={
        "auth": auth_group,
        "config": config_group,
        "secret": secret_group,
        "install": install_cmd,
        "quota": quota_cmd,
        "api-status": status_cmd,
        "honeytoken": honeytoken_group,
        "hmsl": hmsl_group,
    },
)
@click.option(
    "-c",
    "--config-path",
    type=RealPath(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
    is_eager=True,
    help="Set a custom config file. Ignores local and global config files.",
)
@add_common_options()
@click.version_option(version=__version__)
@click.pass_context
def cli(
    ctx: click.Context,
    *,
    allow_self_signed: Optional[bool],
    config_path: Optional[Path],
    **kwargs: Any,
) -> None:
    # Create ContextObj, load config
    ctx.obj = ctx_obj = ContextObj()
    ctx_obj.cache = Cache()
    ctx_obj.config = Config(config_path)
    user_config = ctx_obj.config.user_config

    # If the config wants a higher UI level, set it now
    if user_config.debug and ui.get_level() < ui.Level.DEBUG:
        setup_debug_mode()
    elif user_config.verbose and ui.get_level() < ui.Level.VERBOSE:
        ensure_level(ui.Level.VERBOSE)

    # Update allow_self_signed in the config
    # TODO: this should be reworked: if a command which writes the config is called with
    # --allow-self-signed, the config will contain `allow_self_signed: true`.
    if allow_self_signed:
        user_config.allow_self_signed = allow_self_signed

    ctx_obj.config._dotenv_vars = load_dot_env()

    _set_color(ctx)


def _set_color(ctx: click.Context):
    """
    Helper function to override the default click default output color setting.
        If NO_COLOR is set, we disable color output (see https://no-color.org/).
    If we are in a CI environment, certain variables are set, and we enable colors for
    the logs.
    """
    ci_env_vars = [
        "CI",  # Often set to indicate a generic CI environment
        "GITLAB_CI",
        "GITHUB_ACTIONS",
        "TRAVIS",
        "JENKINS_HOME",
        "JENKINS_URL",
        "CIRCLECI",
        "BITBUCKET_COMMIT",
        "DRONE",
        "BUILD_BUILDID",  # Azure Pipelines
    ]

    if os.getenv("NO_COLOR"):
        ctx.color = False
    elif any(os.getenv(env) for env in ci_env_vars):
        ctx.color = True


def _display_deprecation_message(cfg: Config) -> None:
    for message in cfg.user_config.deprecation_messages:
        ui.display_warning(message)


def _check_for_updates(check_for_updates: bool) -> None:
    # Check for PYTEST_CURRENT_TEST to ensure update check does not happen when running
    # tests: we don't want it to happen because on the CI the unit test-suite is run
    # with --disable-socket, which causes failure on any network access.
    if check_for_updates and "PYTEST_CURRENT_TEST" not in os.environ:
        latest_version = check_updates.check_for_updates()
        if latest_version:
            ui.display_warning(
                f"A new version of ggshield (v{latest_version}) has been released "
                f"(https://github.com/GitGuardian/ggshield)."
            )


@cli.result_callback()
@click.pass_context
def before_exit(ctx: click.Context, exit_code: int, *args: Any, **kwargs: Any) -> None:
    """
    This function is launched as a final callback once subcommands have run.
    It executes some final functions and then terminates.
    The argument exit_code is the result of the previously executed click command.
    """
    ctx_obj = ContextObj.get(ctx)
    _display_deprecation_message(ctx_obj.config)
    _check_for_updates(ctx_obj.check_for_updates)
    sys.exit(exit_code)


def force_utf8_output():
    """
    Force stdout and stderr to always be UTF-8. This is not the case on Windows
    when stdout or stderr is not the console. Doing this fixes integration with
    Visual Studio (see #170).
    """
    for out in sys.stdout, sys.stderr:
        # pyright is not sure sys.stdout and stderr are TextIOWrapper, so it complains when
        # calling `reconfigure()` on them, unless this check is there.
        assert isinstance(out, TextIOWrapper)
        out.reconfigure(encoding="utf-8")


def setup_truststore():
    """Use the system certificates instead of the ones bundled by certifi"""
    if sys.version_info < (3, 10):
        # truststore requires Python 3.10
        return

    import truststore

    truststore.inject_into_ssl()


def main(args: Optional[List[str]] = None) -> Any:
    """
    Wrapper around cli.main() to handle the GITGUARDIAN_CRASH_LOG variable.

    `args` is only used by unit-tests.
    """

    # Required by pyinstaller when forking.
    # See https://pyinstaller.org/en/latest/common-issues-and-pitfalls.html#multi-processing
    multiprocessing.freeze_support()

    log_utils.disable_logs()

    if not os.getenv("GG_PLAINTEXT_OUTPUT", False) and sys.stderr.isatty():
        ui.set_ui(RichGGShieldUI())

    force_utf8_output()
    setup_truststore()

    show_crash_log = getenv_bool("GITGUARDIAN_CRASH_LOG")
    return cli.main(args, prog_name="ggshield", standalone_mode=not show_crash_log)


if __name__ == "__main__":
    sys.exit(main())



import os
import subprocess
from pathlib import Path
from typing import Optional

from ggshield.core import ui
from ggshield.core.errors import UnexpectedError
from ggshield.utils.git_shell import (
    EMPTY_SHA,
    get_last_commit_sha_of_branch,
    get_list_commit_SHA,
    get_new_branch_ci_commits,
    git,
    is_valid_git_commit_ref,
)

from .supported_ci import SupportedCI


def get_previous_commit_from_ci_env() -> Optional[str]:
    """
    Returns the previous HEAD sha of the targeted branch.
    Returns None if there was no commit before.
    """
    supported_ci = SupportedCI.from_ci_env()
    try:
        fcn = PREVIOUS_COMMIT_SHA_FUNCTIONS[supported_ci]
    except KeyError:
        raise UnexpectedError(f"Not implemented for {supported_ci.value}")

    return fcn()


def github_previous_commit_sha() -> Optional[str]:
    push_before_sha = github_push_previous_commit_sha()
    pull_req_base_sha = github_pull_request_previous_commit_sha()
    head_sha = os.getenv("GITHUB_SHA", "HEAD")
    event_name = os.getenv("GITHUB_EVENT_NAME")

    ui.display_verbose(
        f"github_push_before_sha: {push_before_sha}\n"
        f"github_pull_base_sha: {pull_req_base_sha}\n"
    )

    # The PR base sha has to be checked before the push_before_sha
    # because the first one is only populated in case of PR
    # whereas push_before_sha can be populated in PR in case of
    # push force event in a PR
    if pull_req_base_sha:
        return pull_req_base_sha

    if push_before_sha and push_before_sha != EMPTY_SHA:
        return get_commit_except_forced_push(push_before_sha)

    if head_sha and event_name == "push":
        # New branch pushed, try to get sha from git state
        ref_type = os.getenv("GITHUB_REF_TYPE")
        ref_name = os.getenv("GITHUB_REF_NAME")
        if ref_type == "branch" and ref_name:
            return get_new_branch_parent_commit(ref_name)

        ui.display_verbose("Could not find previous commit for current branch.")
        ui.display_verbose("Current branch may have been just pushed.")
        ui.display_verbose("Only scan last commit.")
        last_commits = get_list_commit_SHA(f"{head_sha}~1", max_count=1)
        if len(last_commits) == 1:
            return last_commits[0]

    raise UnexpectedError(
        "Unable to get previous commit. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"github_push_before_sha: {push_before_sha}\n"
        f"github_pull_base_sha: {pull_req_base_sha}\n"
    )


def github_push_previous_commit_sha() -> Optional[str]:
    push_before_sha = os.getenv("GITHUB_PUSH_BEFORE_SHA")

    if not push_before_sha:
        return None

    return push_before_sha


def github_pull_request_previous_commit_sha() -> Optional[str]:
    targeted_branch = os.getenv("GITHUB_BASE_REF")

    # Not in a pull request workflow
    if targeted_branch is None:
        return None

    return get_last_commit_sha_of_branch(f"remotes/origin/{targeted_branch}")


def gitlab_previous_commit_sha() -> Optional[str]:
    push_before_sha = gitlab_push_previous_commit_sha()
    merge_req_base_sha = gitlab_merge_request_previous_commit_sha()
    is_merge_req = bool(os.getenv("CI_MERGE_REQUEST_TARGET_BRANCH_NAME"))

    ui.display_verbose(
        f"gitlab_push_before_sha: {push_before_sha}\n"
        f"gitlab_merge_base_sha: {merge_req_base_sha}\n"
    )

    # push_before_sha is always EMPTY_SHA in MR pipeline according with
    # https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
    if (push_before_sha == EMPTY_SHA or is_merge_req) and merge_req_base_sha:
        # Targeted branch is empty
        if merge_req_base_sha == EMPTY_SHA:
            return None
        return merge_req_base_sha

    if push_before_sha and push_before_sha != EMPTY_SHA:
        return get_commit_except_forced_push(push_before_sha)

    # push_before_sha is also always EMPTY_SHA for the first commit of a new branch
    current_branch = os.getenv("CI_COMMIT_BRANCH")
    if current_branch:
        return get_new_branch_parent_commit(current_branch)

    raise UnexpectedError(
        "Unable to get previous commit. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"gitlab_push_before_sha: {push_before_sha}\n"
        f"gitlab_pull_base_sha: {merge_req_base_sha}\n"
    )


def gitlab_push_previous_commit_sha() -> Optional[str]:
    return os.getenv("CI_COMMIT_BEFORE_SHA")


def gitlab_merge_request_previous_commit_sha() -> Optional[str]:
    targeted_branch = os.getenv("CI_MERGE_REQUEST_TARGET_BRANCH_NAME")

    # Not in a pull request workflow
    if targeted_branch is None:
        return None

    # Requires to fetch targeted branch to access current HEAD
    git(["fetch", "origin", targeted_branch])

    try:
        last_commit = get_last_commit_sha_of_branch(f"origin/{targeted_branch}")
    except Exception:
        # If fail to get last commit of target branch, fallback on CI env variable
        # "CI_MERGE_REQUEST_DIFF_BASE_SHA"
        # This is not the current state of the target branch but the initial state
        # of current branch
        ui.display_verbose(f"Failed to get {targeted_branch} HEAD.")
        ui.display_verbose(
            f"Fallback on commit {os.getenv('CI_MERGE_REQUEST_DIFF_BASE_SHA')}"
        )
        return os.getenv("CI_MERGE_REQUEST_DIFF_BASE_SHA")

    return last_commit


def jenkins_previous_commit_sha() -> Optional[str]:
    previous_commit = os.getenv("GIT_PREVIOUS_COMMIT")
    target_branch = os.getenv("CHANGE_TARGET")
    current_commit = os.getenv("GIT_COMMIT")
    current_branch = os.getenv("GIT_BRANCH")

    ui.display_verbose(
        f"GIT_PREVIOUS_COMMIT: {previous_commit}\n" f"CHANGE_TARGET: {target_branch}\n"
    )

    if target_branch:
        return get_last_commit_sha_of_branch(f"origin/{target_branch}")

    if previous_commit and previous_commit != EMPTY_SHA:
        return get_commit_except_forced_push(previous_commit)

    if current_branch:
        try:
            return get_new_branch_parent_commit(current_branch)
        except subprocess.CalledProcessError:
            ui.display_error("Failed to retrieve initial commit from git history.")

    if current_commit:
        ui.display_verbose("Could not find previous commit for current branch.")
        ui.display_verbose("Current branch may have been just pushed.")
        ui.display_verbose("Only scan last commit.")
        last_commits = get_list_commit_SHA(f"{current_commit}~1", max_count=1)
        if len(last_commits) == 1:
            return last_commits[0]

    raise UnexpectedError(
        "Unable to get previous commit. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"GIT_PREVIOUS_COMMIT: {previous_commit}\n"
        f"CHANGE_TARGET: {target_branch}\n"
    )


def azure_previous_commit_sha() -> Optional[str]:
    push_before_sha = azure_push_previous_commit_sha()
    pull_req_base_sha = azure_pull_request_previous_commit_sha()

    ui.display_verbose(
        f"PUSH_PREVIOUS_COMMIT: {push_before_sha}\n"
        f"PR_TARGET_COMMIT: {pull_req_base_sha}\n"
    )

    if pull_req_base_sha is not None:
        return pull_req_base_sha

    if push_before_sha is not None and push_before_sha != EMPTY_SHA:
        ui.display_verbose(
            "The number of commits of a push event is not available in Azure pipelines."
        )
        ui.display_verbose("Scanning only last commit.")
        return get_commit_except_forced_push(push_before_sha)

    # New branch push
    current_branch = os.getenv("BUILD_SOURCEBRANCHNAME")
    if current_branch:
        return get_new_branch_parent_commit(current_branch)

    # Can't find previous SHA, return last commit
    head_commit = os.getenv("BUILD_SOURCEVERSION")
    last_commits = get_list_commit_SHA(f"{head_commit}~1", max_count=1)

    if len(last_commits) == 0:
        ui.display_verbose("Unable to find commit HEAD~1.")
        return None

    ui.display_verbose(
        "The number of commits of a push event is not available in Azure pipelines."
    )
    ui.display_verbose("Scanning only last commit.")

    if last_commits[0] is not None:
        return last_commits[0]

    raise UnexpectedError(
        "Unable to get previous commit. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"azure_push_before_sha: {push_before_sha}\n"
        f"azure_pull_req_base_sha: {pull_req_base_sha}\n"
    )


def azure_push_previous_commit_sha() -> Optional[str]:
    push_before_sha = os.getenv("BUILD_SOURCEVERSION")

    if not push_before_sha:
        return None

    return push_before_sha


def azure_pull_request_previous_commit_sha() -> Optional[str]:
    """
    Returns commit to compare with in a PR environment in Azure.
    If env variable SYSTEM_PULLREQUEST_TARGETBRANCHNAME is not defined, not in a PR.
    Else, fetch target branch and get commit sha of its HEAD.
    If fail to get this commit sha, returns previous commit in current branch HEAD~1.
    """
    targeted_branch = os.getenv("SYSTEM_PULLREQUEST_TARGETBRANCHNAME")

    # Not in a PR CI job
    if targeted_branch is None:
        return None

    # Requires to fetch targeted branch to access current HEAD
    git(["fetch", "origin", targeted_branch])

    last_commit = get_last_commit_sha_of_branch(f"origin/{targeted_branch}")
    if last_commit is not None:
        return last_commit

    ui.display_verbose("Unable to find HEAD commit sha of target branch.")
    ui.display_verbose("Scans only last commit of current branch.")

    last_commit = get_last_commit_sha_of_branch("HEAD~1")
    if last_commit is not None:
        return last_commit

    ui.display_verbose("Unable to find commit HEAD~1.")
    return None


def get_new_branch_parent_commit(branch_name: str) -> Optional[str]:
    """
    Assuming `branch_name` is a branch newly pushed, this returns a reference
    to the parent commit of all commits pushed on that branch.
    """
    new_commits = get_new_branch_ci_commits(branch_name, Path.cwd(), "origin")
    # New branch pushed with no commits
    if not new_commits:
        ref = f"refs/remotes/origin/{branch_name}"
        return ref if is_valid_git_commit_ref(ref) else "HEAD"
    new_branch_before_sha = f"{new_commits[-1]}^1"
    ui.display_verbose(f"new_branch_before_sha: {new_branch_before_sha}\n")
    if is_valid_git_commit_ref(new_branch_before_sha):
        return new_branch_before_sha
    else:
        ui.display_verbose("> This might be a new repository.")
        return None


def get_commit_except_forced_push(commit: str) -> Optional[str]:
    if not is_valid_git_commit_ref(commit):
        ui.display_verbose("> This might be a forced push.")
        return None
    return commit


PREVIOUS_COMMIT_SHA_FUNCTIONS = {
    SupportedCI.GITLAB: gitlab_previous_commit_sha,
    SupportedCI.GITHUB: github_previous_commit_sha,
    SupportedCI.JENKINS: jenkins_previous_commit_sha,
    SupportedCI.AZURE: azure_previous_commit_sha,
}



from pathlib import Path

from click.testing import CliRunner

from ggshield.__main__ import cli
from tests.unit.conftest import assert_invoke_ok, my_vcr


@my_vcr.use_cassette
def test_hmsl_quota(cli_fs_runner: CliRunner, tmp_path: Path) -> None:
    """
    GIVEN our cli
    WHEN running the hmsl quota command
    THEN we should have a valid response
    """
    result = cli_fs_runner.invoke(cli, ["hmsl", "quota"])
    assert_invoke_ok(result)
    assert "Quota limit" in result.output



# snapshottest: v1 - https://goo.gl/zC4yUc

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_leak_message[_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY_SCAN_RESULT-clip_long_lines-hide_secrets] 1"
] = """> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +1 @@
  1 | +Facebook = 294*********575 | ce3f9f********************5e4372;
                  |__client_id__|
  1 | +Facebook = 294*********575 | ce3f9f********************5e4372;
                                    |_________client_secret________|
"""

snapshots[
    "test_leak_message[_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY_SCAN_RESULT-clip_long_lines-show_secrets] 1"
] = """> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +1 @@
  1 | +Facebook = 294790898041575 | ce3f9f0362bbe5ab01dfc8ee565e4372;
                  |__client_id__|
  1 | +Facebook = 294790898041575 | ce3f9f0362bbe5ab01dfc8ee565e4372;
                                    |_________client_secret________|
"""

snapshots[
    "test_leak_message[_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY_SCAN_RESULT-verbose-hide_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +1 @@
  1 | +Facebook = 294*********575 | ce3f9f********************5e4372;
                  |__client_id__|
  1 | +Facebook = 294*********575 | ce3f9f********************5e4372;
                                    |_________client_secret________|
"""

snapshots[
    "test_leak_message[_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY_SCAN_RESULT-verbose-show_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +1 @@
  1 | +Facebook = 294790898041575 | ce3f9f0362bbe5ab01dfc8ee565e4372;
                  |__client_id__|
  1 | +Facebook = 294790898041575 | ce3f9f0362bbe5ab01dfc8ee565e4372;
                                    |_________client_secret________|
"""

snapshots[
    "test_leak_message[_MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT-clip_long_lines-hide_secrets] 1"
] = """> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +1 @@
  1 | +FacebookAppId = 294*********575; FacebookAppSecret = ce3f9f*************…
                       |__client_id__|
  1 | +…= 294*********575; FacebookAppSecret = ce3f9f********************5e4372;
                                               |_________client_secret________|
"""

snapshots[
    "test_leak_message[_MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT-clip_long_lines-show_secrets] 1"
] = """> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +1 @@
  1 | +FacebookAppId = 294790898041575; FacebookAppSecret = ce3f9f0362bbe5ab01d…
                       |__client_id__|
  1 | +…= 294790898041575; FacebookAppSecret = ce3f9f0362bbe5ab01dfc8ee565e4372;
                                               |_________client_secret________|
"""

snapshots[
    "test_leak_message[_MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT-verbose-hide_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +1 @@
  1 | +FacebookAppId = 294*********575; FacebookAppSecret = ce3f9f********************5e4372;
                       |__client_id__|
  1 | +FacebookAppId = 294*********575; FacebookAppSecret = ce3f9f********************5e4372;
                                                            |_________client_secret________|
"""

snapshots[
    "test_leak_message[_MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT-verbose-show_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +1 @@
  1 | +FacebookAppId = 294790898041575; FacebookAppSecret = ce3f9f0362bbe5ab01dfc8ee565e4372;
                       |__client_id__|
  1 | +FacebookAppId = 294790898041575; FacebookAppSecret = ce3f9f0362bbe5ab01dfc8ee565e4372;
                                                            |_________client_secret________|
"""

snapshots[
    "test_leak_message[_MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT-clip_long_lines-hide_secrets] 1"
] = """> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +2 @@
  1 | +FacebookAppId = 294*********575;
                       |__client_id__|
  2 | +FacebookAppSecret = ce3f9f********************5e4372;
                           |_________client_secret________|
"""

snapshots[
    "test_leak_message[_MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT-clip_long_lines-show_secrets] 1"
] = """> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +2 @@
  1 | +FacebookAppId = 294790898041575;
                       |__client_id__|
  2 | +FacebookAppSecret = ce3f9f0362bbe5ab01dfc8ee565e4372;
                           |_________client_secret________|
"""

snapshots[
    "test_leak_message[_MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT-verbose-hide_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +2 @@
  1 | +FacebookAppId = 294*********575;
                       |__client_id__|
  2 | +FacebookAppSecret = ce3f9f********************5e4372;
                           |_________client_secret________|
"""

snapshots[
    "test_leak_message[_MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT-verbose-show_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 38d9d3464520ed68f18d16e640a4a8b37ef5b17608b455267d100aa487ead314

    | @@ -0,0 +2 @@
  1 | +FacebookAppId = 294790898041575;
                       |__client_id__|
  2 | +FacebookAppSecret = ce3f9f0362bbe5ab01dfc8ee565e4372;
                           |_________client_secret________|
"""

snapshots[
    "test_leak_message[_ONE_LINE_AND_MULTILINE_PATCH_CONTENT-clip_long_lines-hide_secrets] 1"
] = """> This is an example header
> leak.txt: 3 secrets detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 1945f4a0c42abb19c1a420ddd09b4b4681249a3057c427b95f794b18595e7ffa

    | @@ -0,0 +1,29 @@
  1 | +FacebookAppKeys: 294*********573 / ce3f9f********************5e4371 ----…
                        |__client_id__|
  1 | +… 294*********573 / ce3f9f********************5e4371 -----BEGIN RSA PRIV…
                           |_________client_secret________|
  2 | +MIIBOgIBAAJBAIIRkYjxjE3KIZi******************************+******
  3 | +****************************************************************

>> Secret detected: RSA Private Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 060bf63de122848f5efa122fe6cea504aae3b24cea393d887fdefa1529c6a02e

    | @@ -0,0 +1,29 @@
  1 | +…**573 / ce3f9f********************5e4371 -----BEGIN RSA PRIVATE KEY-----
  2 | +MIIBOgIBAAJBAIIRkYjxjE3KIZi******************************+******
  3 | +****************************************************************
  4 | +****************************************************************
  5 | +***********+****************************************************
  6 | +****************+***********************************************
  7 | +**********************+*****************************************
  8 | +****+******Xme/ovcDeM1+3W/UmSHYUW4b3WYq4
  9 | +-----END RSA PRIVATE KEY----- token: SG._Yytrtvlj***********************…
      |_________________________________apikey_________________________________|

>> Secret detected: SendGrid Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 530e5a4a7ea00814db8845dd0cae5efaa4b974a3ce1c76d0384ba715248a5dc1

  7 | +**********************+*****************************************
  8 | +****+******Xme/ovcDeM1+3W/UmSHYUW4b3WYq4
  9 | +…-- token: SG._Yytrtvlj******************************************-**rRJLGFLBLf0M
                  |_______________________________apikey______________________________|
"""

snapshots[
    "test_leak_message[_ONE_LINE_AND_MULTILINE_PATCH_CONTENT-clip_long_lines-show_secrets] 1"
] = """> This is an example header
> leak.txt: 3 secrets detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 1945f4a0c42abb19c1a420ddd09b4b4681249a3057c427b95f794b18595e7ffa

    | @@ -0,0 +1,29 @@
  1 | +FacebookAppKeys: 294790898041573 / ce3f9f0362bbe5ab01dfc8ee565e4371 ----…
                        |__client_id__|
  1 | +… 294790898041573 / ce3f9f0362bbe5ab01dfc8ee565e4371 -----BEGIN RSA PRIV…
                           |_________client_secret________|
  2 | +MIIBOgIBAAJBAIIRkYjxjE3KIZiEc8k4sWWGNsPYRNE0u0bl5oFVApPLm+uXQ/4l
  3 | +bKO9LFtMiVPy700oMWLScwAN5OAiqVLMvHUCAwEAAQJANLr8nmEWuV6t2hAwhK5I

>> Secret detected: RSA Private Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 060bf63de122848f5efa122fe6cea504aae3b24cea393d887fdefa1529c6a02e

    | @@ -0,0 +1,29 @@
  1 | +…41573 / ce3f9f0362bbe5ab01dfc8ee565e4371 -----BEGIN RSA PRIVATE KEY-----
  2 | +MIIBOgIBAAJBAIIRkYjxjE3KIZiEc8k4sWWGNsPYRNE0u0bl5oFVApPLm+uXQ/4l
  3 | +bKO9LFtMiVPy700oMWLScwAN5OAiqVLMvHUCAwEAAQJANLr8nmEWuV6t2hAwhK5I
  4 | +NNmBkEo4M/xFxEtl9J7LKbE2gtNrlCQiJlPP1EMhwAjDOzQcJ3lgFB28dkqH5rMW
  5 | +TQIhANrCE7O+wlCKe0WJqQ3lYlHG91XWyGVgfExJwBDsAD9LAiEAmDY5OSsH0n2A
  6 | +22tthkAvcN1s66lG+0DztOVJ4QLI2z8CIBPeDGwGpx8pdIicN/5LFuLWbyAcoZaT
  7 | +bLaA/DCNPniBAiA0l//bzg+M3srIhm04xzLdR9Vb9IjPRlkvN074zdKDVwIhAKJb
  8 | +RF3C+CMFb0wXme/ovcDeM1+3W/UmSHYUW4b3WYq4
  9 | +-----END RSA PRIVATE KEY----- token: SG._YytrtvljkWqCrkMa3r5hw.yijiPf2qx…
      |_________________________________apikey_________________________________|

>> Secret detected: SendGrid Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 530e5a4a7ea00814db8845dd0cae5efaa4b974a3ce1c76d0384ba715248a5dc1

  7 | +bLaA/DCNPniBAiA0l//bzg+M3srIhm04xzLdR9Vb9IjPRlkvN074zdKDVwIhAKJb
  8 | +RF3C+CMFb0wXme/ovcDeM1+3W/UmSHYUW4b3WYq4
  9 | +…-- token: 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
                  |_______________________________apikey______________________________|
"""

snapshots[
    "test_leak_message[_ONE_LINE_AND_MULTILINE_PATCH_CONTENT-verbose-hide_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 3 secrets detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 1945f4a0c42abb19c1a420ddd09b4b4681249a3057c427b95f794b18595e7ffa

    | @@ -0,0 +1,29 @@
  1 | +FacebookAppKeys: 294*********573 / ce3f9f********************5e4371 -----BEGIN RSA PRIVATE KEY-----
                        |__client_id__|
  1 | +FacebookAppKeys: 294*********573 / ce3f9f********************5e4371 -----BEGIN RSA PRIVATE KEY-----
                                          |_________client_secret________|
  2 | +MIIBOgIBAAJBAIIRkYjxjE3KIZi******************************+******
  3 | +****************************************************************

>> Secret detected: RSA Private Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 060bf63de122848f5efa122fe6cea504aae3b24cea393d887fdefa1529c6a02e

    | @@ -0,0 +1,29 @@
  1 | +FacebookAppKeys: 294*********573 / ce3f9f********************5e4371 -----BEGIN RSA PRIVATE KEY-----
  2 | +MIIBOgIBAAJBAIIRkYjxjE3KIZi******************************+******
  3 | +****************************************************************
  4 | +****************************************************************
  5 | +***********+****************************************************
  6 | +****************+***********************************************
  7 | +**********************+*****************************************
  8 | +****+******Xme/ovcDeM1+3W/UmSHYUW4b3WYq4
  9 | +-----END RSA PRIVATE KEY----- token: SG._Yytrtvlj******************************************-**rRJLGFLBLf0M
      

>> Secret detected: SendGrid Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 530e5a4a7ea00814db8845dd0cae5efaa4b974a3ce1c76d0384ba715248a5dc1

  7 | +**********************+*****************************************
  8 | +****+******Xme/ovcDeM1+3W/UmSHYUW4b3WYq4
  9 | +-----END RSA PRIVATE KEY----- token: SG._Yytrtvlj******************************************-**rRJLGFLBLf0M
                                            |_______________________________apikey______________________________|
"""

snapshots[
    "test_leak_message[_ONE_LINE_AND_MULTILINE_PATCH_CONTENT-verbose-show_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 3 secrets detected

>> Secret detected: Facebook Access Tokens
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 1945f4a0c42abb19c1a420ddd09b4b4681249a3057c427b95f794b18595e7ffa

    | @@ -0,0 +1,29 @@
  1 | +FacebookAppKeys: 294790898041573 / ce3f9f0362bbe5ab01dfc8ee565e4371 -----BEGIN RSA PRIVATE KEY-----
                        |__client_id__|
  1 | +FacebookAppKeys: 294790898041573 / ce3f9f0362bbe5ab01dfc8ee565e4371 -----BEGIN RSA PRIVATE KEY-----
                                          |_________client_secret________|
  2 | +MIIBOgIBAAJBAIIRkYjxjE3KIZiEc8k4sWWGNsPYRNE0u0bl5oFVApPLm+uXQ/4l
  3 | +bKO9LFtMiVPy700oMWLScwAN5OAiqVLMvHUCAwEAAQJANLr8nmEWuV6t2hAwhK5I

>> Secret detected: RSA Private Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 060bf63de122848f5efa122fe6cea504aae3b24cea393d887fdefa1529c6a02e

    | @@ -0,0 +1,29 @@
  1 | +FacebookAppKeys: 294790898041573 / ce3f9f0362bbe5ab01dfc8ee565e4371 -----BEGIN RSA PRIVATE KEY-----
  2 | +MIIBOgIBAAJBAIIRkYjxjE3KIZiEc8k4sWWGNsPYRNE0u0bl5oFVApPLm+uXQ/4l
  3 | +bKO9LFtMiVPy700oMWLScwAN5OAiqVLMvHUCAwEAAQJANLr8nmEWuV6t2hAwhK5I
  4 | +NNmBkEo4M/xFxEtl9J7LKbE2gtNrlCQiJlPP1EMhwAjDOzQcJ3lgFB28dkqH5rMW
  5 | +TQIhANrCE7O+wlCKe0WJqQ3lYlHG91XWyGVgfExJwBDsAD9LAiEAmDY5OSsH0n2A
  6 | +22tthkAvcN1s66lG+0DztOVJ4QLI2z8CIBPeDGwGpx8pdIicN/5LFuLWbyAcoZaT
  7 | +bLaA/DCNPniBAiA0l//bzg+M3srIhm04xzLdR9Vb9IjPRlkvN074zdKDVwIhAKJb
  8 | +RF3C+CMFb0wXme/ovcDeM1+3W/UmSHYUW4b3WYq4
  9 | +-----END RSA PRIVATE KEY----- token: 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
      

>> Secret detected: SendGrid Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 530e5a4a7ea00814db8845dd0cae5efaa4b974a3ce1c76d0384ba715248a5dc1

  7 | +bLaA/DCNPniBAiA0l//bzg+M3srIhm04xzLdR9Vb9IjPRlkvN074zdKDVwIhAKJb
  8 | +RF3C+CMFb0wXme/ovcDeM1+3W/UmSHYUW4b3WYq4
  9 | +-----END RSA PRIVATE KEY----- token: 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
                                            |_______________________________apikey______________________________|
"""

snapshots[
    "test_leak_message[_SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT-clip_long_lines-hide_secrets] 1"
] = """> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: RSA Private Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 060bf63de122848f5efa122fe6cea504aae3b24cea393d887fdefa1529c6a02e

      | @@ -0,0 +1,29 @@
    1 | +PrivateKeyRsa:
    2 | +- text: -----BEGIN RSA PRIVATE KEY-----
    3 | +MIIBOgIBAAJBAIIRkYjxjE3KIZi******************************+******
    4 | +****************************************************************
    5 | +****************************************************************
    6 | +***********+****************************************************
    7 | +****************+***********************************************
    8 | +**********************+*****************************************
    9 | +****+******Xme/ovcDeM1+3W/UmSHYUW4b3WYq4
   10 | +-----END RSA PRIVATE KEY-----
        |_____________________________apikey____________________________|
"""

snapshots[
    "test_leak_message[_SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT-clip_long_lines-show_secrets] 1"
] = """> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: RSA Private Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 060bf63de122848f5efa122fe6cea504aae3b24cea393d887fdefa1529c6a02e

      | @@ -0,0 +1,29 @@
    1 | +PrivateKeyRsa:
    2 | +- text: -----BEGIN RSA PRIVATE KEY-----
    3 | +MIIBOgIBAAJBAIIRkYjxjE3KIZiEc8k4sWWGNsPYRNE0u0bl5oFVApPLm+uXQ/4l
    4 | +bKO9LFtMiVPy700oMWLScwAN5OAiqVLMvHUCAwEAAQJANLr8nmEWuV6t2hAwhK5I
    5 | +NNmBkEo4M/xFxEtl9J7LKbE2gtNrlCQiJlPP1EMhwAjDOzQcJ3lgFB28dkqH5rMW
    6 | +TQIhANrCE7O+wlCKe0WJqQ3lYlHG91XWyGVgfExJwBDsAD9LAiEAmDY5OSsH0n2A
    7 | +22tthkAvcN1s66lG+0DztOVJ4QLI2z8CIBPeDGwGpx8pdIicN/5LFuLWbyAcoZaT
    8 | +bLaA/DCNPniBAiA0l//bzg+M3srIhm04xzLdR9Vb9IjPRlkvN074zdKDVwIhAKJb
    9 | +RF3C+CMFb0wXme/ovcDeM1+3W/UmSHYUW4b3WYq4
   10 | +-----END RSA PRIVATE KEY-----
        |_____________________________apikey____________________________|
"""

snapshots[
    "test_leak_message[_SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT-verbose-hide_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: RSA Private Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 060bf63de122848f5efa122fe6cea504aae3b24cea393d887fdefa1529c6a02e

      | @@ -0,0 +1,29 @@
    1 | +PrivateKeyRsa:
    2 | +- text: -----BEGIN RSA PRIVATE KEY-----
    3 | +MIIBOgIBAAJBAIIRkYjxjE3KIZi******************************+******
    4 | +****************************************************************
    5 | +****************************************************************
    6 | +***********+****************************************************
    7 | +****************+***********************************************
    8 | +**********************+*****************************************
    9 | +****+******Xme/ovcDeM1+3W/UmSHYUW4b3WYq4
   10 | +-----END RSA PRIVATE KEY-----
        |_____________________________apikey____________________________|
"""

snapshots[
    "test_leak_message[_SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT-verbose-show_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: RSA Private Key
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 060bf63de122848f5efa122fe6cea504aae3b24cea393d887fdefa1529c6a02e

      | @@ -0,0 +1,29 @@
    1 | +PrivateKeyRsa:
    2 | +- text: -----BEGIN RSA PRIVATE KEY-----
    3 | +MIIBOgIBAAJBAIIRkYjxjE3KIZiEc8k4sWWGNsPYRNE0u0bl5oFVApPLm+uXQ/4l
    4 | +bKO9LFtMiVPy700oMWLScwAN5OAiqVLMvHUCAwEAAQJANLr8nmEWuV6t2hAwhK5I
    5 | +NNmBkEo4M/xFxEtl9J7LKbE2gtNrlCQiJlPP1EMhwAjDOzQcJ3lgFB28dkqH5rMW
    6 | +TQIhANrCE7O+wlCKe0WJqQ3lYlHG91XWyGVgfExJwBDsAD9LAiEAmDY5OSsH0n2A
    7 | +22tthkAvcN1s66lG+0DztOVJ4QLI2z8CIBPeDGwGpx8pdIicN/5LFuLWbyAcoZaT
    8 | +bLaA/DCNPniBAiA0l//bzg+M3srIhm04xzLdR9Vb9IjPRlkvN074zdKDVwIhAKJb
    9 | +RF3C+CMFb0wXme/ovcDeM1+3W/UmSHYUW4b3WYq4
   10 | +-----END RSA PRIVATE KEY-----
        |_____________________________apikey____________________________|
"""

snapshots[
    "test_leak_message[_SIMPLE_SECRET_PATCH_SCAN_RESULT-clip_long_lines-hide_secrets] 1"
] = """> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: GitHub Token
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 2b5840babacb6f089ddcce1fe5a56b803f8b1f636c6f44cdbf14b0c77a194c93

    | @@ -0,0 +1 @@
  1 | +github_token: 368ac3e**************************37ddf91
                     |________________apikey________________|
"""

snapshots[
    "test_leak_message[_SIMPLE_SECRET_PATCH_SCAN_RESULT-clip_long_lines-show_secrets] 1"
] = """> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: GitHub Token
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 2b5840babacb6f089ddcce1fe5a56b803f8b1f636c6f44cdbf14b0c77a194c93

    | @@ -0,0 +1 @@
  1 | +github_token: 368ac3edf9e850d1c0ff9d6c526496f8237ddf91
                     |________________apikey________________|
"""

snapshots[
    "test_leak_message[_SIMPLE_SECRET_PATCH_SCAN_RESULT-verbose-hide_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: GitHub Token
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 2b5840babacb6f089ddcce1fe5a56b803f8b1f636c6f44cdbf14b0c77a194c93

    | @@ -0,0 +1 @@
  1 | +github_token: 368ac3e**************************37ddf91
                     |________________apikey________________|
"""

snapshots[
    "test_leak_message[_SIMPLE_SECRET_PATCH_SCAN_RESULT-verbose-show_secrets] 1"
] = """
secrets-engine-version: 3.14.159
> This is an example header
> leak.txt: 1 secret detected

>> Secret detected: GitHub Token
   Occurrences: 1
   Known by GitGuardian dashboard: NO
   Incident URL: N/A
   Secret SHA: 2b5840babacb6f089ddcce1fe5a56b803f8b1f636c6f44cdbf14b0c77a194c93

    | @@ -0,0 +1 @@
  1 | +github_token: 368ac3edf9e850d1c0ff9d6c526496f8237ddf91
                     |________________apikey________________|
"""



from datetime import datetime


def datetime_from_isoformat(text: str) -> datetime:
    """Work around for datetime.isoformat() not supporting ISO dates ending with Z"""
    if text.endswith("Z"):
        text = text[:-1] + "+00:00"
    return datetime.fromisoformat(text)


def get_pretty_date(dt: datetime) -> str:
    """
    convert the given datetime to the format September 1, 2022
    """
    # Don't use %d for the day because it adds a leading 0
    return dt.strftime(f"%B {dt.day}, %Y")



from typing import Callable, List, Optional

import click

from ggshield.cmd.utils.common_options import (
    AnyFunction,
    add_common_options,
    create_config_callback,
    create_ctx_callback,
    exit_zero_option,
    get_config_from_context,
    instance_option,
    json_option,
    text_json_sarif_format_option,
)
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.output_format import OutputFormat
from ggshield.core import ui
from ggshield.core.config.user_config import SecretConfig
from ggshield.core.filter import init_exclusion_regexes
from ggshield.utils.click import RealPath
from ggshield.verticals.secret.output import (
    SecretJSONOutputHandler,
    SecretOutputHandler,
    SecretSARIFOutputHandler,
    SecretTextOutputHandler,
)


IGNORED_DEFAULT_WILDCARDS = [
    "**/.git/*/**/*",  # only keep files in .git/ but not in subdirectories
    "**/.pytest_cache/**/*",
    "**/.mypy_cache/**/*",
    "**/.venv/**/*",
    "**/.eggs/**/*",
    "**/.eggs-info/**/*",
    "**/vendor/**/*",
    "**/vendors/**/*",
    "**/node_modules/**/*",
    "top-1000.txt*",
    "**/*.storyboard*",
    "**/*.xib",
    "**/*.mdx*",
    "**/*.sops",
]


def _get_secret_config(ctx: click.Context) -> SecretConfig:
    return get_config_from_context(ctx).secret


_output_option = click.option(
    "--output",
    "-o",
    type=RealPath(exists=False, resolve_path=True),
    default=None,
    help="Redirect ggshield output to PATH.",
    callback=create_ctx_callback("output"),
)


_show_secrets_option = click.option(
    "--show-secrets",
    is_flag=True,
    default=None,
    help="Show secrets in plaintext instead of hiding them.",
    callback=create_config_callback("secret", "show_secrets"),
)


def _exclude_callback(
    ctx: click.Context, param: click.Parameter, value: Optional[List[str]]
) -> Optional[List[str]]:
    ignored_paths = _get_secret_config(ctx).ignored_paths
    if value is not None:
        ignored_paths.update(value)

    ignored_paths.update(IGNORED_DEFAULT_WILDCARDS)
    ctx_obj = ContextObj.get(ctx)
    ctx_obj.exclusion_regexes = init_exclusion_regexes(ignored_paths)
    return value


_exclude_option = click.option(
    "--exclude",
    default=None,
    help="""
    Do not scan paths that match the specified glob-like patterns.
    """,
    multiple=True,
    callback=_exclude_callback,
    metavar="PATTERNS",
)


_ignore_known_secrets_option = click.option(
    "--ignore-known-secrets",
    is_flag=True,
    # The default value **must** be set to None.
    # Each command or subcommand calls create_config_callback to gather option values.
    # If the option is placed early in the command line, the value may be overridden
    # later on with False if no default is defined.
    default=None,
    help="Ignore secrets already known by GitGuardian dashboard.",
    callback=create_config_callback("secret", "ignore_known_secrets"),
)


def _banlist_detectors_callback(
    ctx: click.Context, param: click.Parameter, value: Optional[List[str]]
) -> Optional[List[str]]:
    if value is not None:
        config = _get_secret_config(ctx)
        config.ignored_detectors.update(value)

    return value


_banlist_detectors_option = click.option(
    "--banlist-detector",
    "-b",
    default=None,
    help="Exclude results from a detector.",
    multiple=True,
    callback=_banlist_detectors_callback,
    metavar="DETECTOR",
)


_with_incident_details_option = click.option(
    "--with-incident-details",
    is_flag=True,
    # The default value **must** be set to None.
    # Each command or subcommand calls create_config_callback to gather option values.
    # If the option is placed early in the command line, the value may be overridden
    # later on with False if no default is defined.
    default=None,
    help="""Display full details about the dashboard incident if one is found (JSON and SARIF formats only). Requires the 'incidents:read' scope.""",  # noqa
    callback=create_config_callback("secret", "with_incident_details"),
)


_all_secrets = click.option(
    "--all-secrets",
    is_flag=True,
    help=("Do not ignore any secret. Possible ignore-reason is shown as well."),
    callback=create_config_callback("secret", "all_secrets"),
    default=None,
)


def add_secret_scan_common_options() -> Callable[[AnyFunction], AnyFunction]:
    def decorator(cmd: AnyFunction) -> AnyFunction:
        add_common_options()(cmd)
        json_option(cmd)
        text_json_sarif_format_option(cmd)
        _output_option(cmd)
        _show_secrets_option(cmd)
        exit_zero_option(cmd)
        _exclude_option(cmd)
        _ignore_known_secrets_option(cmd)
        _banlist_detectors_option(cmd)
        _with_incident_details_option(cmd)
        instance_option(cmd)
        _all_secrets(cmd)
        return cmd

    return decorator


OUTPUT_HANDLER_CLASSES = {
    OutputFormat.TEXT: SecretTextOutputHandler,
    OutputFormat.JSON: SecretJSONOutputHandler,
    OutputFormat.SARIF: SecretSARIFOutputHandler,
}


def create_output_handler(ctx: click.Context) -> SecretOutputHandler:
    """Read objects defined in ctx.obj and create the appropriate OutputHandler
    instance"""
    ctx_obj = ContextObj.get(ctx)
    output_handler_cls = OUTPUT_HANDLER_CLASSES[ctx_obj.output_format]
    config = ctx_obj.config
    return output_handler_cls(
        verbose=ui.is_verbose(),
        client=ctx_obj.client,
        output=ctx_obj.output,
        secret_config=config.user_config.secret,
    )



"""The primitives used for encryption and hashing"""

import secrets
from hashlib import sha256

from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt


# Publicly available pepper used in our protocol with HasMySecretLeaked
_PEPPER = sha256(b"GitGuardian").digest()


class DecryptionError(Exception):
    """Raised when a message couldn't be decrypted,
    for instance because it wasn't encrypted with the same key."""


def hash_string(text: str) -> str:
    return (
        Scrypt(salt=_PEPPER, n=2048, r=8, p=1, length=32)
        .derive(text.encode("utf-8"))
        .hex()
    )


def make_hint(key_as_hex: str) -> str:
    """Make a hint from a key, to be used
    so that users can know which payload to decrypt."""
    return sha256(bytes.fromhex(key_as_hex)).hexdigest()


def encrypt(message: str, key: bytes) -> bytes:
    """Encrypt a message using AES-GCM with a random nonce.

    Key must be 32 bytes long.
    """
    nonce = secrets.token_bytes(12)
    return nonce + AESGCM(key).encrypt(
        nonce=nonce, data=message.encode("utf-8"), associated_data=None
    )


def decrypt(payload: bytes, key: bytes) -> str:
    """Decrypt a message using AES-GCM. Key must be 32 bytes long."""
    try:
        cleartext = AESGCM(key).decrypt(
            nonce=payload[:12], data=payload[12:], associated_data=None
        )
        return cleartext.decode()
    except InvalidTag as exc:
        raise DecryptionError from exc



from pathlib import Path

from tests.conftest import GG_VALID_TOKEN
from tests.functional.utils import run_ggshield_scan


def test_scan_docset_positive(tmp_path: Path) -> None:
    test_file = tmp_path / "docset.jsonl"
    test_file.write_text(
        '{"type": "github_issue", "id": "https://github.com/foo/bar/issues/2", "documents": ['
        + '{"id": "https://github.com/foo/bar/issues/2", "content": "issue", "authors": [{"id": "ghost"}]}]}\n'
        + '{"type": "github_issue", "id": "https://github.com/foo/bar/issues/1", "documents": ['
        + '{"id": "https://github.com/foo/bar/issues/1", "content": "Token in comment", "authors": [{"id": "ghost"}]},'
        + '{"id": "https://github.com/foo/bar/issues/1#issuecomment-42",'
        + f'"content": "apikey = {GG_VALID_TOKEN}",'
        + ' "authors": [{"id": "ghost2"}]}'
        + "]}"
    )

    result = run_ggshield_scan("docset", str(test_file), cwd=tmp_path, expected_code=1)
    assert "apikey =" in result.stdout
    assert (
        "https://github.com/foo/bar/issues/1#issuecomment-42: 1 secret detected"
        in result.stdout
    )


def test_scan_docset_negative(tmp_path: Path) -> None:
    test_file = tmp_path / "docset.jsonl"
    test_file.write_text(
        '{"type": "github_issue", "id": "https://github.com/foo/bar/issues/2", "documents": ['
        + '{"id": "https://github.com/foo/bar/issues/2", "content": "issue", "authors": [{"id": "ghost"}]}]}\n'
        + '{"type": "github_issue", "id": "https://github.com/foo/bar/issues/1", "documents": ['
        + '{"id": "https://github.com/foo/bar/issues/1", "content": "No token", "authors": [{"id": "ghost"}]},'
        + '{"id": "https://github.com/foo/bar/issues/1#issuecomment-42",'
        + '"content": "Nothing to see, move along",'
        + ' "authors": [{"id": "ghost2"}]}'
        + "]}"
    )

    run_ggshield_scan("docset", str(test_file), cwd=tmp_path, expected_code=0)



from pathlib import Path

import pytest

from tests.functional.utils import run_ggshield_scan


@pytest.mark.parametrize(
    "package, expected_code",
    (
        ("ggshield==1.14.2", 1),  # ggshield 1.14.2 contains some test secrets
        ("marshmallow", 0),
    ),
)
def test_scan_pypi(tmp_path: Path, package: str, expected_code: int) -> None:
    # Run the command from a temporary path to ensure we don't load the
    # .gitguardian.yaml stored at the root of ggshield repo.
    # If we did then we would not find secrets in ggshield package because the
    # test secrets would be ignored.
    run_ggshield_scan("pypi", package, expected_code=expected_code, cwd=tmp_path)



import logging
import os
import re
from enum import Enum
from pathlib import Path
from typing import Any, Optional, Set, Tuple

import click

from ggshield.core.config.auth_config import AuthConfig
from ggshield.core.config.user_config import UserConfig
from ggshield.core.config.utils import remove_url_trailing_slash
from ggshield.core.constants import DEFAULT_HMSL_URL, DEFAULT_INSTANCE_URL
from ggshield.core.url_utils import (
    api_to_dashboard_url,
    clean_url,
    dashboard_to_api_url,
    validate_instance_url,
)


class ConfigSource(Enum):
    """
    Enum of the different sources of configuration
    where an API key or instance URL can come from
    """

    CMD_OPTION = "command line option"
    DOTENV = ".env file"
    ENV_VAR = "environment variable"
    USER_CONFIG = "user config"
    DEFAULT = "default"


logger = logging.getLogger(__name__)


class Config:
    """
    Top-level config class. Contains an instance of UserConfig and an instance of
    AuthConfig.
    """

    __slots__ = [
        "user_config",
        "auth_config",
        "_cmdline_instance_name",
        "_config_path",
        "_dotenv_vars",
    ]

    user_config: UserConfig
    auth_config: AuthConfig

    # The instance name, if ggshield is invoked with `--instance`
    _cmdline_instance_name: Optional[str]

    _config_path: Path

    # This environment variable helps us knowing whether environment variables
    # were set by the dotenv file or not
    # It is used in the `api-status` command to return the API key and instance sources
    _dotenv_vars: Set[str]

    def __init__(self, config_path: Optional[Path] = None):
        self.user_config, self._config_path = UserConfig.load(config_path=config_path)
        self.auth_config = AuthConfig.load()
        self._cmdline_instance_name = None
        self._dotenv_vars = set()

    def save(self) -> None:
        self.user_config.save(self._config_path)
        self.auth_config.save()

    @property
    def config_path(self) -> Path:
        return self._config_path

    @property
    def instance_name(self) -> str:
        return self.get_instance_name_and_source()[0]

    def get_instance_name_and_source(self) -> Tuple[str, ConfigSource]:
        """
        Return the instance name and source of the selected instance.

        The instance name (defaulting to URL) of the selected instance
        priority order is:
        - set from the command line (by setting cmdline_instance_name)
          - in case the user set the api url instead of dashboard url, we replace it
        - GITGUARDIAN_INSTANCE env var
        - GITGUARDIAN_API_URL env var
        - in local user config (in user_config.dashboard_url)
        - in global user config (in user_config.dashboard_url)
        - the default instance
        """
        if self._cmdline_instance_name:
            if re.match(
                r"^https:\/\/api(\.[a-z0-9]+)?\.gitguardian\.com",
                self._cmdline_instance_name,
            ) or re.match(r"/exposed/?$", self._cmdline_instance_name):
                return (
                    api_to_dashboard_url(self._cmdline_instance_name),
                    ConfigSource.CMD_OPTION,
                )
            return self._cmdline_instance_name, ConfigSource.CMD_OPTION

        try:
            url = os.environ["GITGUARDIAN_INSTANCE"]
            logger.debug("Using instance URL from $GITGUARDIAN_INSTANCE")
        except KeyError:
            pass
        else:
            validate_instance_url(url)
            source = (
                ConfigSource.DOTENV
                if "GITGUARDIAN_INSTANCE" in self._dotenv_vars
                else ConfigSource.ENV_VAR
            )
            return remove_url_trailing_slash(url), source

        try:
            name = os.environ["GITGUARDIAN_API_URL"]
            logger.debug("Using API URL from $GITGUARDIAN_API_URL")
        except KeyError:
            pass
        else:
            source = (
                ConfigSource.DOTENV
                if "GITGUARDIAN_API_URL" in self._dotenv_vars
                else ConfigSource.ENV_VAR
            )
            return api_to_dashboard_url(name, warn=True), source

        if self.user_config.instance:
            return self.user_config.instance, ConfigSource.USER_CONFIG

        return DEFAULT_INSTANCE_URL, ConfigSource.DEFAULT

    @property
    def cmdline_instance_name(self) -> Optional[str]:
        return self._cmdline_instance_name

    @cmdline_instance_name.setter
    def cmdline_instance_name(self, name: str) -> None:
        """
        Override the current instance name. To be called by commands supporting the
        `--instance` option.
        """
        parsed_url = clean_url(name)
        if not parsed_url.scheme or not parsed_url.netloc:
            raise click.BadParameter(
                "Please provide a valid URL.",
                param_hint="instance",
            )
        self._cmdline_instance_name = f"{parsed_url.scheme}://{parsed_url.netloc}"

    @property
    def api_url(self) -> str:
        """
        The API URL to use the API
        It's the API URL from the configured instance
        """
        # TODO change when instance_name can be a name instead of just a URL
        return dashboard_to_api_url(self.instance_name)

    @property
    def dashboard_url(self) -> str:
        """
        The dashboard URL to use the dashboard
        It's the dashboard URL from the configured instance
        """
        # TODO change when instance_name can be a name instead of just a URL
        return self.instance_name

    @property
    def api_key(self) -> str:
        return self.get_api_key_and_source()[0]

    def get_api_key_and_source(self) -> Tuple[str, ConfigSource]:
        """
        Return the selected API key and its source

        The API key to use
        priority order is
        - env var
        - the api key from the selected instance
        """
        try:
            key = os.environ["GITGUARDIAN_API_KEY"]
            logger.debug("Using API key from $GITGUARDIAN_API_KEY")
            source = (
                ConfigSource.DOTENV
                if "GITGUARDIAN_API_KEY" in self._dotenv_vars
                else ConfigSource.ENV_VAR
            )
        except KeyError:
            key = self.auth_config.get_instance_token(self.instance_name)
            source = ConfigSource.USER_CONFIG
        return key, source

    def add_ignored_match(self, *args: Any, **kwargs: Any) -> None:
        return self.user_config.secret.add_ignored_match(*args, **kwargs)

    @property
    def saas_api_url(self) -> str:
        """
        The GIM SaaS instance used to get JWT tokens.
        This can be overridden with the GITGUARDIAN_SAAS_URL environment variable.
        If the HMSL URL is different from the default, we derive the SaaS API from it.
        Otherwise the SaaS API URL is actually the one derived from the dashboard
        """
        if self.hmsl_url != DEFAULT_HMSL_URL and "api" in self.hmsl_url:
            default_value = self.hmsl_url.replace("hasmysecretleaked", "gitguardian")
        else:
            default_value = self.api_url

        return os.environ.get(
            "GITGUARDIAN_SAAS_URL",
            default_value,
        )

    @property
    def saas_api_key(self) -> str:
        """
        The API key associated with the SaaS GIM instance.
        For testing purposes, it can be overridden with the
        GITGUARDIAN_SAAS_API_KEY environment variable.
        """
        try:
            key = os.environ["GITGUARDIAN_SAAS_API_KEY"]
            logger.debug("Using API key from $GITGUARDIAN_SAAS_API_KEY")
        except KeyError:
            pass
        else:
            return key
        logger.debug("Using API key for SaaS instance from config")
        return self.api_key

    # Properties for HasMySecretLeaked
    # We can't rely on the instance selected by the user,
    # as JWT creation is only available in SaaS.
    @property
    def hmsl_url(self) -> str:
        """
        The url of the HasMySecretLeaked service
        """
        try:
            url = os.environ["GITGUARDIAN_HMSL_URL"]
            logger.debug("Using HasMySecretLeaked URL from $GITGUARDIAN_HMSL_URL")
        except KeyError:
            url = DEFAULT_HMSL_URL
            logger.debug("Using default HasMySecretLeaked URL")
        return remove_url_trailing_slash(url)

    @property
    def hmsl_audience(self) -> str:
        """
        The audience of our JWT tokens for HasMySecretLeaked.
        """
        try:
            audience = os.environ["GITGUARDIAN_HMSL_AUDIENCE"]
            logger.debug("Using audience from $GITGUARDIAN_HMSL_AUDIENCE")
        except KeyError:
            audience = self.hmsl_url
            logger.debug("Using HasMySecretLeaked URL as audience")
        return audience



from pathlib import Path
from typing import List, Pattern, Set, Tuple, Union

from ggshield.utils.files import ListFilesMode, is_path_binary, list_files, url_for_path

from .scannable import Scannable


class File(Scannable):
    """Implementation of Scannable for files from the disk."""

    def __init__(self, path: Union[str, Path]):
        super().__init__()
        self._path = Path(path)

    @property
    def url(self) -> str:
        return url_for_path(self._path)

    @property
    def filename(self) -> str:
        return str(self._path)

    @property
    def path(self) -> Path:
        return self._path

    def is_longer_than(self, max_utf8_encoded_size: int) -> bool:
        if self._utf8_encoded_size is not None:
            # We already have the encoded size, easy
            return self._utf8_encoded_size > max_utf8_encoded_size

        with self.path.open("rb") as fp:
            (
                result,
                self._content,
                self._utf8_encoded_size,
            ) = Scannable._is_file_longer_than(fp, max_utf8_encoded_size)
        return result

    def _read_content(self) -> None:
        if self._content is None:
            with self.path.open("rb") as f:
                self._content, self._utf8_encoded_size = Scannable._decode_bytes(
                    f.read()
                )


def create_files_from_paths(
    paths: List[Path],
    exclusion_regexes: Set[Pattern[str]],
    list_files_mode: ListFilesMode = ListFilesMode.GIT_COMMITTED_OR_STAGED,
) -> Tuple[List[Scannable], List[Path]]:
    """
    Create File instances for `paths` and return them, as well as a list of the ignored
    paths found in `paths`.
    """
    filepaths = list_files(
        paths,
        exclusion_regexes,
        list_files_mode=list_files_mode,
    )

    files: List[Scannable] = []
    binary_paths: List[Path] = []
    for path in filepaths:
        if is_path_binary(path):
            binary_paths.append(path)
            continue

        files.append(File(path))

    return (files, binary_paths)



from marshmallow import fields
from pygitguardian.models import SecretIncidentSchema
from pygitguardian.models_utils import BaseSchema

from ggshield.verticals.secret.extended_match import ExtendedMatchSchema


class IgnoreReasonSchema(BaseSchema):
    kind = fields.String(required=True)
    detail = fields.String()


class FlattenedPolicyBreak(BaseSchema):
    policy = fields.String(required=True)
    occurrences = fields.List(fields.Nested(ExtendedMatchSchema), required=True)
    detector = fields.String(data_key="type", required=True)
    validity = fields.String(required=False, allow_none=True)
    ignore_sha = fields.String(required=True)
    total_occurrences = fields.Integer(required=True)
    incident_url = fields.String(required=True, dump_default="")
    incident_details = fields.Nested(SecretIncidentSchema)
    known_secret = fields.Bool(required=True, dump_default=False)
    ignore_reason = fields.Nested(IgnoreReasonSchema, dump_default=None)


class JSONResultSchema(BaseSchema):
    mode = fields.String(required=True)
    filename = fields.String(required=True)
    incidents = fields.List(fields.Nested(FlattenedPolicyBreak), required=True)
    total_incidents = fields.Integer(required=True)
    total_occurrences = fields.Integer(required=True)


class JSONErrorSchema(BaseSchema):
    class JSONErrorFileSchema(BaseSchema):
        mode = fields.String(required=True)
        filename = fields.String(required=True)

    files = fields.List(fields.Nested(JSONErrorFileSchema))
    description = fields.String(required=True)


class JSONScanCollectionSchema(BaseSchema):
    id = fields.String()
    type = fields.String()
    results = fields.List(
        fields.Nested(JSONResultSchema), data_key="entities_with_incidents"
    )
    errors = fields.List(fields.Nested(JSONErrorSchema))
    scans = fields.List(fields.Nested(lambda: JSONScanCollectionSchema()))
    extra_info = fields.Dict(keys=fields.Str(), values=fields.Str())
    total_incidents = fields.Integer(required=True)
    total_occurrences = fields.Integer(required=True)
    secrets_engine_version = fields.String(required=False)



import json
from typing import Any, Dict, Iterator, TextIO, cast

import click

from ggshield.cmd.hmsl.hmsl_common_options import input_arg
from ggshield.cmd.utils.common_options import (
    add_common_options,
    json_option,
    text_json_format_option,
)
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.errors import ParseError
from ggshield.verticals.hmsl import Match, Secret
from ggshield.verticals.hmsl.crypto import make_hint
from ggshield.verticals.hmsl.output import show_results


@click.command()
@click.pass_context
@add_common_options()
@click.option(
    "--mapping",
    "-m",
    "mapping_file",
    type=click.File("r"),
    default="mapping.txt",
    show_default=True,
    help="File containing the hashes and their names.",
)
@text_json_format_option
@json_option
@input_arg
def decrypt_cmd(ctx: click.Context, path: str, mapping_file: TextIO, **_: Any) -> int:
    """
    Decrypt `query`'s output and show secrets information.
    """
    ctx_obj = ContextObj.get(ctx)

    # Opens the file or stdin
    input = cast(TextIO, click.open_file(path, "r"))
    mapping: Dict[str, str] = load_mapping(mapping_file)

    # Decrypt the secrets thanks to the hashes contained in mapping
    try:
        secrets = decrypt(input, mapping)
        # Display the secrets
        show_results(secrets, mapping, ctx_obj.use_json)
    except (json.JSONDecodeError, TypeError):
        raise ParseError("Invalid format in input file.")

    return 0


# Helper methods


def load_mapping(mapping_file: TextIO) -> Dict[str, str]:
    """
    Load the mapping from the file.
    """
    mapping: Dict[str, str] = {}
    for line in mapping_file:
        line = line.strip()
        if line == "":
            # Skip empty lines
            continue
        hash, name = line.split(":", maxsplit=1)
        mapping[hash] = name
    return mapping


def decrypt(input: TextIO, mapping: Dict[str, str]) -> Iterator[Secret]:
    """
    Decrypt the secrets thanks to the hashes contained in mapping.
    """
    hashes = set(mapping.keys())
    hints = {make_hint(hash): hash for hash in hashes}
    for line in input:
        line = line.strip()
        if line == "":
            # Skip empty lines
            continue
        data = json.loads(line)
        if data.keys() == {"hint", "payload"}:
            match = Match(**json.loads(line))
            if match.hint in hints:
                yield match.decrypt(hints[match.hint])
        else:
            yield Secret(**data)



import json
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import click

from ggshield.cmd.utils.common_options import (
    add_common_options,
    json_option,
    text_json_format_option,
)
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.config.auth_config import InstanceConfig

from .constants import DATETIME_FORMAT, FIELDS


@dataclass
class InstanceInfo:
    instance_name: str
    default_token_lifetime: Optional[int]
    workspace_id: Any
    url: str
    token: str
    token_name: str
    expiry: str


@dataclass
class ConfigData:
    instances: List[InstanceInfo] = field(default_factory=list)
    global_values: Dict[str, Any] = field(default_factory=dict)

    def as_dict(self) -> Dict[str, Any]:
        return {
            "instances": [instance.__dict__ for instance in self.instances],
            "global_values": self.global_values,
        }


def get_instance_info(
    instance: InstanceConfig, default_token_lifetime: Any
) -> InstanceInfo:
    """Helper function to extract instance information."""
    instance_name = instance.name or instance.url
    account = instance.account

    if account is not None:
        workspace_id = account.workspace_id
        token = account.token
        token_name = account.token_name
        expire_at = account.expire_at
        expiry = expire_at.strftime(DATETIME_FORMAT) if expire_at else "never"
    else:
        workspace_id = token = token_name = expiry = "not set"

    _default_token_lifetime = instance.default_token_lifetime or default_token_lifetime

    return InstanceInfo(
        instance_name=instance_name,
        default_token_lifetime=_default_token_lifetime,
        workspace_id=workspace_id,
        url=instance.url,
        token=token,
        token_name=token_name,
        expiry=expiry,
    )


@click.command()
@click.pass_context
@json_option
@text_json_format_option
@add_common_options()
def config_list_cmd(ctx: click.Context, **kwargs: Any) -> int:
    """
    Print the list of configuration keys and values.
    """
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    default_token_lifetime = config.auth_config.default_token_lifetime

    config_data = ConfigData()
    for config_field in FIELDS.values():
        config_obj = (
            config.auth_config if config_field.auth_config else config.user_config
        )
        value = getattr(config_obj, config_field.name)
        config_data.global_values[config_field.name] = value

    config_data.instances = [
        get_instance_info(instance, default_token_lifetime)
        for instance in config.auth_config.instances
    ]

    if ctx_obj.use_json:
        click.echo(json.dumps(config_data.as_dict()))
    else:
        message_lines = [
            f"{key}: {value}" for key, value in config_data.global_values.items()
        ]
        message_lines.append("")
        for instance in config_data.instances:
            message_lines.append(f"[{instance.instance_name}]")
            for key, value in instance.__dict__.items():
                if key != "instance_name":
                    message_lines.append(f"{key}: {value}")
            message_lines.append("")

        click.echo("\n".join(message_lines).strip())

    return 0
